import unittest
import itertools

from dbt.exceptions import VersionsNotCompatibleException
from dbt.semver import VersionSpecifier, UnboundedVersionSpecifier, \
    VersionRange, reduce_versions, versions_compatible, \
    resolve_to_specific_version


def create_range(start_version_string, end_version_string):
    start = UnboundedVersionSpecifier()
    end = UnboundedVersionSpecifier()

    if start_version_string is not None:
        start = VersionSpecifier.from_version_string(start_version_string)

    if end_version_string is not None:
        end = VersionSpecifier.from_version_string(end_version_string)

    return VersionRange(start=start, end=end)


class TestSemver(unittest.TestCase):

    def assertVersionSetResult(self, inputs, output_range):
        expected = create_range(*output_range)

        for permutation in itertools.permutations(inputs):
            self.assertDictEqual(
                reduce_versions(*permutation),
                expected)

    def assertInvalidVersionSet(self, inputs):
        for permutation in itertools.permutations(inputs):
            with self.assertRaises(VersionsNotCompatibleException):
                reduce_versions(*permutation)

    def test__versions_compatible(self):
        self.assertTrue(
            versions_compatible('0.0.1', '0.0.1'))
        self.assertFalse(
            versions_compatible('0.0.1', '0.0.2'))
        self.assertTrue(
            versions_compatible('>0.0.1', '0.0.2'))

    def test__reduce_versions(self):
        self.assertVersionSetResult(
            ['0.0.1', '0.0.1'],
            ['=0.0.1', '=0.0.1'])

        self.assertVersionSetResult(
            ['0.0.1'],
            ['=0.0.1', '=0.0.1'])

        self.assertVersionSetResult(
            ['>0.0.1'],
            ['>0.0.1', None])

        self.assertVersionSetResult(
            ['<0.0.1'],
            [None, '<0.0.1'])

        self.assertVersionSetResult(
            ['>0.0.1', '0.0.2'],
            ['=0.0.2', '=0.0.2'])

        self.assertVersionSetResult(
            ['0.0.2', '>=0.0.2'],
            ['=0.0.2', '=0.0.2'])

        self.assertVersionSetResult(
            ['>0.0.1', '>0.0.2', '>0.0.3'],
            ['>0.0.3', None])

        self.assertVersionSetResult(
            ['>0.0.1', '<0.0.3'],
            ['>0.0.1', '<0.0.3'])

        self.assertVersionSetResult(
            ['>0.0.1', '0.0.2', '<0.0.3'],
            ['=0.0.2', '=0.0.2'])

        self.assertVersionSetResult(
            ['>0.0.1', '>=0.0.1', '<0.0.3'],
            ['>0.0.1', '<0.0.3'])

        self.assertVersionSetResult(
            ['>0.0.1', '<0.0.3', '<=0.0.3'],
            ['>0.0.1', '<0.0.3'])

        self.assertVersionSetResult(
            ['>0.0.1', '>0.0.2', '<0.0.3', '<0.0.4'],
            ['>0.0.2', '<0.0.3'])

        self.assertVersionSetResult(
            ['<=0.0.3', '>=0.0.3'],
            ['>=0.0.3', '<=0.0.3'])

        self.assertInvalidVersionSet(['>0.0.2', '0.0.1'])
        self.assertInvalidVersionSet(['>0.0.2', '0.0.2'])
        self.assertInvalidVersionSet(['<0.0.2', '0.0.2'])
        self.assertInvalidVersionSet(['<0.0.2', '>0.0.3'])
        self.assertInvalidVersionSet(['<=0.0.3', '>0.0.3'])
        self.assertInvalidVersionSet(['<0.0.3', '>=0.0.3'])
        self.assertInvalidVersionSet(['<0.0.3', '>0.0.3'])

    def test__resolve_to_specific_version(self):
        self.assertEqual(
            resolve_to_specific_version(
                create_range('>0.0.1', None),
                ['0.0.1', '0.0.2']),
            '0.0.2')

        self.assertEqual(
            resolve_to_specific_version(
                create_range('>=0.0.2', None),
                ['0.0.1', '0.0.2']),
            '0.0.2')

        self.assertEqual(
            resolve_to_specific_version(
                create_range('>=0.0.3', None),
                ['0.0.1', '0.0.2']),
            None)

        self.assertEqual(
            resolve_to_specific_version(
                create_range('>=0.0.3', '<0.0.5'),
                ['0.0.3', '0.0.4', '0.0.5']),
            '0.0.4')

        self.assertEqual(
            resolve_to_specific_version(
                create_range(None, '<=0.0.5'),
                ['0.0.3', '0.1.4', '0.0.5']),
            '0.0.5')



import os
import unittest
from unittest.mock import MagicMock, patch

import dbt.clients.system
import dbt.compilation
import dbt.context.parser
import dbt.exceptions
import dbt.flags
import dbt.linker
import dbt.parser
import dbt.config
import dbt.utils
import dbt.parser.manifest
from dbt.contracts.graph.manifest import FilePath, SourceFile, FileHash
from dbt.parser.results import ParseResult
from dbt.parser.base import BaseParser

try:
    from queue import Empty
except ImportError:
    from Queue import Empty

from dbt.logger import GLOBAL_LOGGER as logger # noqa

from .utils import config_from_parts_or_dicts


class GraphTest(unittest.TestCase):

    def tearDown(self):
        self.write_gpickle_patcher.stop()
        self.load_projects_patcher.stop()
        self.file_system_patcher.stop()
        self.get_adapter_patcher.stop()
        self.get_adapter_patcher_cmn.stop()
        self.mock_filesystem_constructor.stop()
        self.mock_hook_constructor.stop()
        self.load_patch.stop()
        self.load_source_file_patcher.stop()

    def setUp(self):
        dbt.flags.STRICT_MODE = True
        self.graph_result = None

        self.write_gpickle_patcher = patch('networkx.write_gpickle')
        self.load_projects_patcher = patch('dbt.parser.manifest._load_projects')
        self.file_system_patcher = patch.object(
            dbt.parser.search.FilesystemSearcher, '__new__'
        )
        self.hook_patcher = patch.object(
            dbt.parser.hooks.HookParser, '__new__'
        )
        self.get_adapter_patcher = patch('dbt.context.parser.get_adapter')
        self.factory = self.get_adapter_patcher.start()
        # also patch this one

        self.get_adapter_patcher_cmn = patch('dbt.context.common.get_adapter')
        self.factory_cmn = self.get_adapter_patcher_cmn.start()


        def mock_write_gpickle(graph, outfile):
            self.graph_result = graph
        self.mock_write_gpickle = self.write_gpickle_patcher.start()
        self.mock_write_gpickle.side_effect = mock_write_gpickle

        self.profile = {
            'outputs': {
                'test': {
                    'type': 'postgres',
                    'threads': 4,
                    'host': 'thishostshouldnotexist',
                    'port': 5432,
                    'user': 'root',
                    'pass': 'password',
                    'dbname': 'dbt',
                    'schema': 'dbt_test'
                }
            },
            'target': 'test'
        }

        self.mock_load_projects = self.load_projects_patcher.start()
        def _load_projects(config, paths):
            yield config.project_name, config
        self.mock_load_projects.side_effect = _load_projects

        self.mock_models = []

        def _mock_parse_result(config, all_projects):
            return ParseResult(
                vars_hash=FileHash.from_contents('vars'),
                project_hashes={name: FileHash.from_contents(name) for name in all_projects},
                profile_hash=FileHash.from_contents('profile'),
            )

        self.load_patch = patch('dbt.parser.manifest.make_parse_result')
        self.mock_parse_result = self.load_patch.start()
        self.mock_parse_result.side_effect = _mock_parse_result

        self.load_source_file_patcher = patch.object(BaseParser, 'load_file')
        self.mock_source_file = self.load_source_file_patcher.start()
        self.mock_source_file.side_effect = lambda path: [n for n in self.mock_models if n.path == path][0]

        def filesystem_iter(iter_self):
            if 'sql' not in iter_self.extension:
                return []
            if 'models' not in iter_self.relative_dirs:
                return []
            return [model.path for model in self.mock_models]

        def create_filesystem_searcher(cls, project, relative_dirs, extension):
            result = MagicMock(project=project, relative_dirs=relative_dirs, extension=extension)
            result.__iter__.side_effect = lambda: iter(filesystem_iter(result))
            return result

        def create_hook_patcher(cls, results, project, relative_dirs, extension):
            result = MagicMock(results=results, project=project, relative_dirs=relative_dirs, extension=extension)
            result.__iter__.side_effect = lambda: iter([])
            return result

        self.mock_filesystem_constructor = self.file_system_patcher.start()
        self.mock_filesystem_constructor.side_effect = create_filesystem_searcher
        self.mock_hook_constructor = self.hook_patcher.start()
        self.mock_hook_constructor.side_effect = create_hook_patcher

    def get_config(self, extra_cfg=None):
        if extra_cfg is None:
            extra_cfg = {}

        cfg = {
            'name': 'test_models_compile',
            'version': '0.1',
            'profile': 'test',
            'project-root': os.path.abspath('.'),
        }
        cfg.update(extra_cfg)

        return config_from_parts_or_dicts(project=cfg, profile=self.profile)

    def get_compiler(self, project):
        return dbt.compilation.Compiler(project)

    def use_models(self, models):
        for k, v in models.items():
            path = FilePath(
                searched_path='models',
                project_root=os.path.normcase(os.getcwd()),
                relative_path='{}.sql'.format(k),
            )
            source_file = SourceFile(path=path, checksum=FileHash.empty())
            source_file.contents = v
            self.mock_models.append(source_file)

    def load_manifest(self, config):
        loader = dbt.parser.manifest.ManifestLoader(config, {config.project_name: config})
        loader.load()
        return loader.create_manifest()

    def test__single_model(self):
        self.use_models({
            'model_one': 'select * from events',
        })

        config = self.get_config()
        manifest = self.load_manifest(config)

        compiler = self.get_compiler(config)
        linker = compiler.compile(manifest)

        self.assertEqual(
            list(linker.nodes()),
            ['model.test_models_compile.model_one'])

        self.assertEqual(
            list(linker.edges()),
            [])

    def test__two_models_simple_ref(self):
        self.use_models({
            'model_one': 'select * from events',
            'model_two': "select * from {{ref('model_one')}}",
        })

        config = self.get_config()
        manifest = self.load_manifest(config)
        compiler = self.get_compiler(config)
        linker = compiler.compile(manifest)

        self.assertCountEqual(
            linker.nodes(),
            [
                'model.test_models_compile.model_one',
                'model.test_models_compile.model_two',
            ]
        )

        self.assertCountEqual(
            linker.edges(),
            [('model.test_models_compile.model_one', 'model.test_models_compile.model_two',)]
        )

    def test__model_materializations(self):
        self.use_models({
            'model_one': 'select * from events',
            'model_two': "select * from {{ref('model_one')}}",
            'model_three': "select * from events",
            'model_four': "select * from events",
        })

        cfg = {
            "models": {
                "materialized": "table",
                "test_models_compile": {
                    "model_one": {"materialized": "table"},
                    "model_two": {"materialized": "view"},
                    "model_three": {"materialized": "ephemeral"}
                }
            }
        }

        config = self.get_config(cfg)
        manifest = self.load_manifest(config)
        compiler = self.get_compiler(config)
        linker = compiler.compile(manifest)

        expected_materialization = {
            "model_one": "table",
            "model_two": "view",
            "model_three": "ephemeral",
            "model_four": "table"
        }

        for model, expected in expected_materialization.items():
            key = 'model.test_models_compile.{}'.format(model)
            actual = manifest.nodes[key].config.materialized
            self.assertEqual(actual, expected)

    def test__model_incremental(self):
        self.use_models({
            'model_one': 'select * from events'
        })

        cfg = {
            "models": {
                "test_models_compile": {
                    "model_one": {
                        "materialized": "incremental",
                        "unique_key": "id"
                    },
                }
            }
        }

        config = self.get_config(cfg)
        manifest = self.load_manifest(config)
        compiler = self.get_compiler(config)
        linker = compiler.compile(manifest)

        node = 'model.test_models_compile.model_one'

        self.assertEqual(list(linker.nodes()), [node])
        self.assertEqual(list(linker.edges()), [])

        self.assertEqual(manifest.nodes[node].config.materialized, 'incremental')

    def test__dependency_list(self):
        self.use_models({
            'model_1': 'select * from events',
            'model_2': 'select * from {{ ref("model_1") }}',
            'model_3': '''
                select * from {{ ref("model_1") }}
                union all
                select * from {{ ref("model_2") }}
            ''',
            'model_4': 'select * from {{ ref("model_3") }}'
        })

        config = self.get_config()
        manifest = self.load_manifest(config)
        compiler = self.get_compiler(config)
        linker = compiler.compile(manifest)

        models = ('model_1', 'model_2', 'model_3', 'model_4')
        model_ids = ['model.test_models_compile.{}'.format(m) for m in models]

        manifest = MagicMock(nodes={
            n: MagicMock(unique_id=n)
            for n in model_ids
        })
        manifest.expect.side_effect = lambda n: MagicMock(unique_id=n)
        queue = linker.as_graph_queue(manifest)

        for model_id in model_ids:
            self.assertFalse(queue.empty())
            got = queue.get(block=False)
            self.assertEqual(got.unique_id, model_id)
            with self.assertRaises(Empty):
                queue.get(block=False)
            queue.mark_done(got.unique_id)
        self.assertTrue(queue.empty())

    def test__partial_parse(self):
        config = self.get_config()

        loader = dbt.parser.manifest.ManifestLoader(config, {config.project_name: config})
        loader.load()
        loader.create_manifest()
        results = loader.results

        self.assertTrue(loader.matching_parse_results(results))
        too_low = results.replace(dbt_version='0.0.1a1')
        self.assertFalse(loader.matching_parse_results(too_low))
        too_high = results.replace(dbt_version='99999.99.99')
        self.assertFalse(loader.matching_parse_results(too_high))



import pickle
from datetime import timedelta

from dbt.contracts.graph.unparsed import (
    UnparsedNode, UnparsedRunHook, UnparsedMacro, Time, TimePeriod,
    FreshnessStatus, FreshnessThreshold, Quoting, UnparsedSourceDefinition,
    UnparsedSourceTableDefinition, UnparsedDocumentationFile, NamedTested,
    UnparsedNodeUpdate
)
from dbt.node_types import NodeType
from .utils import ContractTestCase


class TestUnparsedMacro(ContractTestCase):
    ContractType = UnparsedMacro

    def test_ok(self):
        macro_dict = {
            'path': '/root/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': '{% macro foo() %}select 1 as id{% endmacro %}',
            'root_path': '/root/',
            'resource_type': 'macro',
        }
        macro = self.ContractType(
            path='/root/path.sql',
            original_file_path='/root/path.sql',
            package_name='test',
            raw_sql='{% macro foo() %}select 1 as id{% endmacro %}',
            root_path='/root/',
            resource_type=NodeType.Macro,
        )
        self.assert_symmetric(macro, macro_dict)
        pickle.loads(pickle.dumps(macro))

    def test_invalid_missing_field(self):
        macro_dict = {
            'path': '/root/path.sql',
            'original_file_path': '/root/path.sql',
            # 'package_name': 'test',
            'raw_sql': '{% macro foo() %}select 1 as id{% endmacro %}',
            'root_path': '/root/',
            'resource_type': 'macro',
        }
        self.assert_fails_validation(macro_dict)

    def test_invalid_extra_field(self):
        macro_dict = {
            'path': '/root/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': '{% macro foo() %}select 1 as id{% endmacro %}',
            'root_path': '/root/',
            'extra': 'extra',
            'resource_type': 'macro',
        }
        self.assert_fails_validation(macro_dict)


class TestUnparsedNode(ContractTestCase):
    ContractType = UnparsedNode

    def test_ok(self):
        node_dict = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': NodeType.Model,
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from {{ ref("thing") }}',
        }
        node = self.ContractType(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from {{ ref("thing") }}',
            name='foo',
            resource_type=NodeType.Model,
        )
        self.assert_symmetric(node, node_dict)
        self.assertFalse(node.empty)

        self.assert_fails_validation(node_dict, cls=UnparsedRunHook)
        self.assert_fails_validation(node_dict, cls=UnparsedMacro)
        pickle.loads(pickle.dumps(node))

    def test_empty(self):
        node_dict = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': NodeType.Model,
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': '  \n',
        }
        node = UnparsedNode(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='  \n',
            name='foo',
            resource_type=NodeType.Model,
        )
        self.assert_symmetric(node, node_dict)
        self.assertTrue(node.empty)

        self.assert_fails_validation(node_dict, cls=UnparsedRunHook)
        self.assert_fails_validation(node_dict, cls=UnparsedMacro)

    def test_bad_type(self):
        node_dict = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': NodeType.Source,  # not valid!
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from {{ ref("thing") }}',
        }
        self.assert_fails_validation(node_dict)


class TestUnparsedRunHook(ContractTestCase):
    ContractType = UnparsedRunHook

    def test_ok(self):
        node_dict = {
            'name': 'foo',
            'root_path': 'test/dbt_project.yml',
            'resource_type': NodeType.Operation,
            'path': '/root/dbt_project.yml',
            'original_file_path': '/root/dbt_project.yml',
            'package_name': 'test',
            'raw_sql': 'GRANT select on dbt_postgres',
            'index': 4
        }
        node = self.ContractType(
            package_name='test',
            root_path='test/dbt_project.yml',
            path='/root/dbt_project.yml',
            original_file_path='/root/dbt_project.yml',
            raw_sql='GRANT select on dbt_postgres',
            name='foo',
            resource_type=NodeType.Operation,
            index=4,
        )
        self.assert_symmetric(node, node_dict)
        self.assert_fails_validation(node_dict, cls=UnparsedNode)
        pickle.loads(pickle.dumps(node))

    def test_bad_type(self):
        node_dict = {
            'name': 'foo',
            'root_path': 'test/dbt_project.yml',
            'resource_type': NodeType.Model,  # invalid
            'path': '/root/dbt_project.yml',
            'original_file_path': '/root/dbt_project.yml',
            'package_name': 'test',
            'raw_sql': 'GRANT select on dbt_postgres',
            'index': 4
        }
        self.assert_fails_validation(node_dict)


class TestFreshnessThreshold(ContractTestCase):
    ContractType = FreshnessThreshold

    def test_empty(self):
        empty = self.ContractType(None, None)
        self.assert_symmetric(empty, {})
        self.assertEqual(empty.status(float('Inf')), FreshnessStatus.Pass)
        self.assertEqual(empty.status(0), FreshnessStatus.Pass)

    def test_both(self):
        threshold = self.ContractType(
            warn_after=Time(count=18, period=TimePeriod.hour),
            error_after=Time(count=2, period=TimePeriod.day),
        )
        dct = {
            'error_after': {'count': 2, 'period': 'day'},
            'warn_after': {'count': 18, 'period': 'hour'}
        }
        self.assert_symmetric(threshold, dct)

        error_seconds = timedelta(days=3).total_seconds()
        warn_seconds = timedelta(days=1).total_seconds()
        pass_seconds = timedelta(hours=3).total_seconds()
        self.assertEqual(threshold.status(error_seconds), FreshnessStatus.Error)
        self.assertEqual(threshold.status(warn_seconds), FreshnessStatus.Warn)
        self.assertEqual(threshold.status(pass_seconds), FreshnessStatus.Pass)
        pickle.loads(pickle.dumps(threshold))

    def test_merged(self):
        t1 = self.ContractType(
            warn_after=Time(count=36, period=TimePeriod.hour),
            error_after=Time(count=2, period=TimePeriod.day),
        )
        t2 = self.ContractType(
            warn_after=Time(count=18, period=TimePeriod.hour),
        )
        threshold = self.ContractType(
            warn_after=Time(count=18, period=TimePeriod.hour),
            error_after=Time(count=2, period=TimePeriod.day),
        )
        self.assertEqual(threshold, t1.merged(t2))

        error_seconds = timedelta(days=3).total_seconds()
        warn_seconds = timedelta(days=1).total_seconds()
        pass_seconds = timedelta(hours=3).total_seconds()
        self.assertEqual(threshold.status(error_seconds), FreshnessStatus.Error)
        self.assertEqual(threshold.status(warn_seconds), FreshnessStatus.Warn)
        self.assertEqual(threshold.status(pass_seconds), FreshnessStatus.Pass)


class TestQuoting(ContractTestCase):
    ContractType = Quoting

    def test_empty(self):
        empty = self.ContractType()
        self.assert_symmetric(empty, {})

    def test_partial(self):
        a = self.ContractType(None, True, False)
        b = self.ContractType(True, False, None)
        self.assert_symmetric(a, {'schema': True, 'identifier': False})
        self.assert_symmetric(b, {'database': True, 'schema': False})

        c = a.merged(b)
        self.assertEqual(c, self.ContractType(True, False, False))
        self.assert_symmetric(
            c, {'database': True, 'schema': False, 'identifier': False}
        )
        pickle.loads(pickle.dumps(c))


class TestUnparsedSourceDefinition(ContractTestCase):
    ContractType = UnparsedSourceDefinition

    def test_defaults(self):
        minimum = self.ContractType(name='foo')
        self.assert_from_dict(minimum, {'name': 'foo'})
        self.assert_to_dict(minimum, {'name': 'foo', 'description': '', 'freshness': {}, 'quoting': {}, 'tables': [], 'loader': ''})

    def test_contents(self):
        empty = self.ContractType(
            name='foo',
            description='a description',
            quoting=Quoting(database=False),
            loader='some_loader',
            freshness=FreshnessThreshold(),
            tables=[],
        )
        dct = {
            'name': 'foo',
            'description': 'a description',
            'quoting': {'database': False},
            'loader': 'some_loader',
            'freshness': {},
            'tables': [],
        }
        self.assert_symmetric(empty, dct)

    def test_table_defaults(self):
        table_1 = UnparsedSourceTableDefinition(name='table1')
        table_2 = UnparsedSourceTableDefinition(
            name='table2',
            description='table 2',
            quoting=Quoting(database=True),
        )
        source = self.ContractType(
            name='foo',
            tables=[table_1, table_2]
        )
        from_dict = {
            'name': 'foo',
            'tables': [
                {'name': 'table1'},
                {
                    'name': 'table2',
                    'description': 'table 2',
                    'quoting': {'database': True},
                },
            ],
        }
        to_dict = {
            'name': 'foo',
            'description': '',
            'loader': '',
            'freshness': {},
            'quoting': {},
            'tables': [
                {
                    'name': 'table1',
                    'description': '',
                    'tests': [],
                    'columns': [],
                    'quoting': {},
                    'external': {},
                    'freshness': {},
                },
                {
                    'name': 'table2',
                    'description': 'table 2',
                    'tests': [],
                    'columns': [],
                    'quoting': {'database': True},
                    'external': {},
                    'freshness': {},
                },
            ],
        }
        self.assert_from_dict(source, from_dict)
        self.assert_symmetric(source, to_dict)
        pickle.loads(pickle.dumps(source))


class TestUnparsedDocumentationFile(ContractTestCase):
    ContractType = UnparsedDocumentationFile

    def test_ok(self):
        doc = self.ContractType(
            package_name='test',
            root_path='/root',
            path='/root/docs',
            original_file_path='/root/docs/doc.md',
            file_contents='blah blah blah',
        )
        doc_dict = {
            'package_name': 'test',
            'root_path': '/root',
            'path': '/root/docs',
            'original_file_path': '/root/docs/doc.md',
            'file_contents': 'blah blah blah',
        }
        self.assert_symmetric(doc, doc_dict)
        self.assertEqual(doc.resource_type, NodeType.Documentation)
        self.assert_fails_validation(doc_dict, UnparsedNode)
        pickle.loads(pickle.dumps(doc))

    def test_extra_field(self):
        self.assert_fails_validation({})
        doc_dict = {
            'package_name': 'test',
            'root_path': '/root',
            'path': '/root/docs',
            'original_file_path': '/root/docs/doc.md',
            'file_contents': 'blah blah blah',
            'resource_type': 'docs',
        }
        self.assert_fails_validation(doc_dict)


class TestUnparsedNodeUpdate(ContractTestCase):
    ContractType = UnparsedNodeUpdate

    def test_defaults(self):
        minimum = self.ContractType(name='foo')
        from_dict = {'name': 'foo'}
        to_dict = {'name': 'foo', 'columns': [], 'description': '', 'tests': []}
        self.assert_from_dict(minimum, from_dict)
        self.assert_to_dict(minimum, to_dict)

    def test_contents(self):
        update = self.ContractType(
            name='foo',
            description='a description',
            tests=['table_test'],
            columns=[
                NamedTested(name='x', description='x description'),
                NamedTested(
                    name='y',
                    description='y description',
                    tests=[
                        'unique',
                        {'accepted_values': {'values': ['blue', 'green']}}
                    ]
                ),
            ],
        )
        dct = {
            'name': 'foo',
            'description': 'a description',
            'tests': ['table_test'],
            'columns': [
                {'name': 'x', 'description': 'x description', 'tests': []},
                {
                    'name': 'y',
                    'description': 'y description',
                    'tests': [
                        'unique',
                        {'accepted_values': {'values': ['blue', 'green']}}
                    ],
                },
            ],
        }
        self.assert_symmetric(update, dct)
        pickle.loads(pickle.dumps(update))

    def test_bad_test_type(self):
        dct = {
            'name': 'foo',
            'description': 'a description',
            'tests': ['table_test'],
            'columns': [
                {'name': 'x', 'description': 'x description', 'tests': []},
                {
                    'name': 'y',
                    'description': 'y description',
                    'tests': [
                        100,
                        {'accepted_values': {'values': ['blue', 'green']}}
                    ],
                },
            ],
        }
        self.assert_fails_validation(dct)

        dct = {
            'name': 'foo',
            'description': 'a description',
            'tests': ['table_test'],
            'columns': [
                # column missing a name
                {'description': 'x description', 'tests': []},
                {
                    'name': 'y',
                    'description': 'y description',
                    'tests': [
                        'unique',
                        {'accepted_values': {'values': ['blue', 'green']}}
                    ],
                },
            ],
        }
        self.assert_fails_validation(dct)

        # missing a name
        dct = {
            'description': 'a description',
            'tests': ['table_test'],
            'columns': [
                {'name': 'x', 'description': 'x description', 'tests': []},
                {
                    'name': 'y',
                    'description': 'y description',
                    'tests': [
                        'unique',
                        {'accepted_values': {'values': ['blue', 'green']}}
                    ],
                },
            ],
        }
        self.assert_fails_validation(dct)



import pickle

from dbt.contracts.graph.compiled import (
    CompiledModelNode, InjectedCTE, CompiledTestNode
)
from dbt.contracts.graph.parsed import (
    DependsOn, NodeConfig, TestConfig
)
from dbt.node_types import NodeType

from .utils import ContractTestCase


class TestCompiledModelNode(ContractTestCase):
    ContractType = CompiledModelNode

    def test_basic_uncompiled(self):
        node_dict = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Model),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': []},
            'database': 'test_db',
            'description': '',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': [],
            'config': {
                'column_types': {},
                'enabled': True,
                'materialized': 'view',
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
            },
            'docrefs': [],
            'columns': {},
            'compiled': False,
            'extra_ctes': [],
            'extra_ctes_injected': False,
        }
        node = self.ContractType(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from wherever',
            name='foo',
            resource_type=NodeType.Model,
            unique_id='model.test.foo',
            fqn=['test', 'models', 'foo'],
            refs=[],
            sources=[],
            depends_on=DependsOn(),
            description='',
            database='test_db',
            schema='test_schema',
            alias='bar',
            tags=[],
            config=NodeConfig(),
            compiled=False,
            extra_ctes=[],
            extra_ctes_injected=False,
        )
        self.assert_symmetric(node, node_dict)
        self.assertFalse(node.empty)
        self.assertTrue(node.is_refable)
        self.assertFalse(node.is_ephemeral)
        self.assertEqual(node.local_vars(), {})

        minimum = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Model),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'database': 'test_db',
            'schema': 'test_schema',
            'alias': 'bar',
        }
        self.assert_from_dict(node, minimum)
        pickle.loads(pickle.dumps(node))

    def test_basic_compiled(self):
        node_dict = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Model),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from {{ ref("other") }}',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': []},
            'database': 'test_db',
            'description': '',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': [],
            'config': {
                'column_types': {},
                'enabled': True,
                'materialized': 'view',
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
            },
            'docrefs': [],
            'columns': {},
            'compiled': True,
            'compiled_sql': 'select * from whatever',
            'extra_ctes': [{'id': 'whatever', 'sql': 'select * from other'}],
            'extra_ctes_injected': True,
            'injected_sql': 'with whatever as (select * from other) select * from whatever',
            'wrapped_sql': 'None',
        }
        node = self.ContractType(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from {{ ref("other") }}',
            name='foo',
            resource_type=NodeType.Model,
            unique_id='model.test.foo',
            fqn=['test', 'models', 'foo'],
            refs=[],
            sources=[],
            depends_on=DependsOn(),
            description='',
            database='test_db',
            schema='test_schema',
            alias='bar',
            tags=[],
            config=NodeConfig(),
            compiled=True,
            compiled_sql='select * from whatever',
            extra_ctes=[InjectedCTE('whatever', 'select * from other')],
            extra_ctes_injected=True,
            injected_sql='with whatever as (select * from other) select * from whatever',
            wrapped_sql='None',
        )
        self.assert_symmetric(node, node_dict)
        self.assertFalse(node.empty)
        self.assertTrue(node.is_refable)
        self.assertFalse(node.is_ephemeral)
        self.assertEqual(node.local_vars(), {})

    def test_invalid_extra_fields(self):
        bad_extra = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Model),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'database': 'test_db',
            'schema': 'test_schema',
            'alias': 'bar',
            'notvalid': 'nope',
        }
        self.assert_fails_validation(bad_extra)

    def test_invalid_bad_type(self):
        bad_type = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Macro),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'database': 'test_db',
            'schema': 'test_schema',
            'alias': 'bar',
        }
        self.assert_fails_validation(bad_type)


class TestCompiledTestNode(ContractTestCase):
    ContractType = CompiledTestNode

    def test_basic_uncompiled(self):
        node_dict = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Test),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': []},
            'database': 'test_db',
            'description': '',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': [],
            'config': {
                'column_types': {},
                'enabled': True,
                'materialized': 'view',
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
                'severity': 'error',
            },
            'docrefs': [],
            'columns': {},
            'compiled': False,
            'extra_ctes': [],
            'extra_ctes_injected': False,
        }
        node = self.ContractType(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from wherever',
            name='foo',
            resource_type=NodeType.Test,
            unique_id='model.test.foo',
            fqn=['test', 'models', 'foo'],
            refs=[],
            sources=[],
            depends_on=DependsOn(),
            description='',
            database='test_db',
            schema='test_schema',
            alias='bar',
            tags=[],
            config=TestConfig(),
            compiled=False,
            extra_ctes=[],
            extra_ctes_injected=False,
        )
        self.assert_symmetric(node, node_dict)
        self.assertFalse(node.empty)
        self.assertFalse(node.is_refable)
        self.assertFalse(node.is_ephemeral)
        self.assertEqual(node.local_vars(), {})

        minimum = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Test),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'database': 'test_db',
            'schema': 'test_schema',
            'alias': 'bar',
        }
        self.assert_from_dict(node, minimum)
        pickle.loads(pickle.dumps(node))

    def test_basic_compiled(self):
        node_dict = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Test),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from {{ ref("other") }}',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'refs': [],
            'sources': [],
            'depends_on': {'macros': [], 'nodes': []},
            'database': 'test_db',
            'description': '',
            'schema': 'test_schema',
            'alias': 'bar',
            'tags': [],
            'config': {
                'column_types': {},
                'enabled': True,
                'materialized': 'view',
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'quoting': {},
                'tags': [],
                'vars': {},
                'severity': 'warn',
            },
            'docrefs': [],
            'columns': {},
            'compiled': True,
            'compiled_sql': 'select * from whatever',
            'extra_ctes': [{'id': 'whatever', 'sql': 'select * from other'}],
            'extra_ctes_injected': True,
            'injected_sql': 'with whatever as (select * from other) select * from whatever',
            'wrapped_sql': 'select count(*) from (with whatever as (select * from other) select * from whatever) sbq',
            'column_name': 'id',
        }
        node = self.ContractType(
            package_name='test',
            root_path='/root/',
            path='/root/x/path.sql',
            original_file_path='/root/path.sql',
            raw_sql='select * from {{ ref("other") }}',
            name='foo',
            resource_type=NodeType.Test,
            unique_id='model.test.foo',
            fqn=['test', 'models', 'foo'],
            refs=[],
            sources=[],
            depends_on=DependsOn(),
            description='',
            database='test_db',
            schema='test_schema',
            alias='bar',
            tags=[],
            config=TestConfig(severity='warn'),
            compiled=True,
            compiled_sql='select * from whatever',
            extra_ctes=[InjectedCTE('whatever', 'select * from other')],
            extra_ctes_injected=True,
            injected_sql='with whatever as (select * from other) select * from whatever',
            wrapped_sql='select count(*) from (with whatever as (select * from other) select * from whatever) sbq',
            column_name='id',
        )
        self.assert_symmetric(node, node_dict)
        self.assertFalse(node.empty)
        self.assertFalse(node.is_refable)
        self.assertFalse(node.is_ephemeral)
        self.assertEqual(node.local_vars(), {})

    def test_invalid_extra_fields(self):
        bad_extra = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Test),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'database': 'test_db',
            'schema': 'test_schema',
            'alias': 'bar',
            'extra': 'extra value',
        }
        self.assert_fails_validation(bad_extra)

    def test_invalid_resource_type(self):
        bad_type = {
            'name': 'foo',
            'root_path': '/root/',
            'resource_type': str(NodeType.Model),
            'path': '/root/x/path.sql',
            'original_file_path': '/root/path.sql',
            'package_name': 'test',
            'raw_sql': 'select * from wherever',
            'unique_id': 'model.test.foo',
            'fqn': ['test', 'models', 'foo'],
            'database': 'test_db',
            'schema': 'test_schema',
            'alias': 'bar',
        }
        self.assert_fails_validation(bad_type)



import unittest
from unittest import mock

import dbt.exceptions
from dbt.task.deps import (
    GitUnpinnedPackage, LocalUnpinnedPackage, RegistryUnpinnedPackage,
    LocalPackageContract, GitPackageContract, RegistryPackageContract,
    resolve_packages
)
from dbt.contracts.project import PackageConfig
from dbt.semver import VersionSpecifier

from hologram import ValidationError


class TestLocalPackage(unittest.TestCase):
    def test_init(self):
        a_contract = LocalPackageContract.from_dict({'local': '/path/to/package'})
        self.assertEqual(a_contract.local, '/path/to/package')
        a = LocalUnpinnedPackage.from_contract(a_contract)
        self.assertEqual(a.local, '/path/to/package')
        a_pinned = a.resolved()
        self.assertEqual(a_pinned.local, '/path/to/package')
        self.assertEqual(str(a_pinned), '/path/to/package')


class TestGitPackage(unittest.TestCase):
    def test_init(self):
        a_contract = GitPackageContract.from_dict(
            {'git': 'http://example.com', 'revision': '0.0.1'}
        )
        self.assertEqual(a_contract.git, 'http://example.com')
        self.assertEqual(a_contract.revision, '0.0.1')
        self.assertIs(a_contract.warn_unpinned, None)

        a = GitUnpinnedPackage.from_contract(a_contract)
        self.assertEqual(a.git, 'http://example.com')
        self.assertEqual(a.revisions, ['0.0.1'])
        self.assertIs(a.warn_unpinned, True)

        a_pinned = a.resolved()
        self.assertEqual(a_pinned.name, 'http://example.com')
        self.assertEqual(a_pinned.get_version(), '0.0.1')
        self.assertEqual(a_pinned.source_type(), 'git')
        self.assertIs(a_pinned.warn_unpinned, True)

    def test_invalid(self):
        with self.assertRaises(ValidationError):
            GitPackageContract.from_dict(
                {'git': 'http://example.com', 'version': '0.0.1'}
            )

    def test_resolve_ok(self):
        a_contract = GitPackageContract.from_dict(
            {'git': 'http://example.com', 'revision': '0.0.1'}
        )
        b_contract = GitPackageContract.from_dict(
            {'git': 'http://example.com', 'revision': '0.0.1',
             'warn-unpinned': False}
        )
        a = GitUnpinnedPackage.from_contract(a_contract)
        b = GitUnpinnedPackage.from_contract(b_contract)
        self.assertTrue(a.warn_unpinned)
        self.assertFalse(b.warn_unpinned)
        c = a.incorporate(b)

        c_pinned = c.resolved()
        self.assertEqual(c_pinned.name, 'http://example.com')
        self.assertEqual(c_pinned.get_version(), '0.0.1')
        self.assertEqual(c_pinned.source_type(), 'git')
        self.assertFalse(c_pinned.warn_unpinned)

    def test_resolve_fail(self):
        a_contract = GitPackageContract.from_dict(
            {'git': 'http://example.com', 'revision': '0.0.1'}
        )
        b_contract = GitPackageContract.from_dict(
            {'git': 'http://example.com', 'revision': '0.0.2'}
        )
        a = GitUnpinnedPackage.from_contract(a_contract)
        b = GitUnpinnedPackage.from_contract(b_contract)
        c = a.incorporate(b)
        self.assertEqual(c.git, 'http://example.com')
        self.assertEqual(c.revisions, ['0.0.1', '0.0.2'])

        with self.assertRaises(dbt.exceptions.DependencyException):
            c.resolved()

    def test_default_revision(self):
        a_contract = GitPackageContract.from_dict({'git': 'http://example.com'})
        self.assertEqual(a_contract.revision, None)
        self.assertIs(a_contract.warn_unpinned, None)

        a = GitUnpinnedPackage.from_contract(a_contract)
        self.assertEqual(a.git, 'http://example.com')
        self.assertEqual(a.revisions, [])
        self.assertIs(a.warn_unpinned, True)

        a_pinned = a.resolved()
        self.assertEqual(a_pinned.name, 'http://example.com')
        self.assertEqual(a_pinned.get_version(), 'master')
        self.assertEqual(a_pinned.source_type(), 'git')
        self.assertIs(a_pinned.warn_unpinned, True)


class TestHubPackage(unittest.TestCase):
    def setUp(self):
        self.patcher = mock.patch('dbt.task.deps.registry')
        self.registry = self.patcher.start()
        self.index_cached = self.registry.index_cached
        self.get_available_versions = self.registry.get_available_versions
        self.package_version = self.registry.package_version

        self.index_cached.return_value = [
            'fishtown-analytics-test/a',
        ]
        self.get_available_versions.return_value = [
            '0.1.2', '0.1.3'
        ]
        self.package_version.return_value = {
            'id': 'fishtown-analytics-test/a/0.1.2',
            'name': 'a',
            'version': '0.1.2',
            'packages': [],
            '_source': {
                'blahblah': 'asdfas',
            },
            'downloads': {
                'tarball': 'https://example.com/invalid-url!',
                'extra': 'field',
            },
            'newfield': ['another', 'value'],
        }

    def tearDown(self):
        self.patcher.stop()

    def test_init(self):
        a_contract = RegistryPackageContract(
            package='fishtown-analytics-test/a',
            version='0.1.2',
        )
        self.assertEqual(a_contract.package, 'fishtown-analytics-test/a')
        self.assertEqual(a_contract.version, '0.1.2')

        a = RegistryUnpinnedPackage.from_contract(a_contract)
        self.assertEqual(a.package, 'fishtown-analytics-test/a')
        self.assertEqual(
            a.versions,
            [VersionSpecifier(
                build=None,
                major='0',
                matcher='=',
                minor='1',
                patch='2',
                prerelease=None
            )]
        )

        a_pinned = a.resolved()
        self.assertEqual(a_contract.package, 'fishtown-analytics-test/a')
        self.assertEqual(a_contract.version, '0.1.2')
        self.assertEqual(a_pinned.source_type(), 'hub')

    def test_invalid(self):
        with self.assertRaises(ValidationError):
            RegistryPackageContract.from_dict(
                {'package': 'namespace/name', 'key': 'invalid'}
            )

    def test_resolve_ok(self):
        a_contract = RegistryPackageContract(
            package='fishtown-analytics-test/a',
            version='0.1.2'
        )
        b_contract = RegistryPackageContract(
            package='fishtown-analytics-test/a',
            version='0.1.2'
        )
        a = RegistryUnpinnedPackage.from_contract(a_contract)
        b = RegistryUnpinnedPackage.from_contract(b_contract)
        c = a.incorporate(b)

        self.assertEqual(c.package, 'fishtown-analytics-test/a')
        self.assertEqual(
            c.versions,
            [
                VersionSpecifier(
                    build=None,
                    major='0',
                    matcher='=',
                    minor='1',
                    patch='2',
                    prerelease=None,
                ),
                VersionSpecifier(
                    build=None,
                    major='0',
                    matcher='=',
                    minor='1',
                    patch='2',
                    prerelease=None,
                ),
            ]
        )

        c_pinned = c.resolved()
        self.assertEqual(c_pinned.package, 'fishtown-analytics-test/a')
        self.assertEqual(c_pinned.version, '0.1.2')
        self.assertEqual(c_pinned.source_type(), 'hub')

    def test_resolve_missing_package(self):
        a = RegistryUnpinnedPackage.from_contract(RegistryPackageContract(
            package='fishtown-analytics-test/b',
            version='0.1.2'
        ))
        with self.assertRaises(dbt.exceptions.DependencyException) as exc:
            a.resolved()

        msg = 'Package fishtown-analytics-test/b was not found in the package index'
        self.assertEqual(msg, str(exc.exception))

    def test_resolve_missing_version(self):
        a = RegistryUnpinnedPackage.from_contract(RegistryPackageContract(
            package='fishtown-analytics-test/a',
            version='0.1.4'
        ))

        with self.assertRaises(dbt.exceptions.DependencyException) as exc:
            a.resolved()
        msg = (
            "Could not find a matching version for package "
            "fishtown-analytics-test/a\n  Requested range: =0.1.4, =0.1.4\n  "
            "Available versions: ['0.1.2', '0.1.3']"
        )
        self.assertEqual(msg, str(exc.exception))

    def test_resolve_conflict(self):
        a_contract = RegistryPackageContract(
            package='fishtown-analytics-test/a',
            version='0.1.2'
        )
        b_contract = RegistryPackageContract(
            package='fishtown-analytics-test/a',
            version='0.1.3'
        )
        a = RegistryUnpinnedPackage.from_contract(a_contract)
        b = RegistryUnpinnedPackage.from_contract(b_contract)
        c = a.incorporate(b)

        with self.assertRaises(dbt.exceptions.DependencyException) as exc:
            c.resolved()
        msg = (
            "Version error for package fishtown-analytics-test/a: Could not "
            "find a satisfactory version from options: ['=0.1.2', '=0.1.3']"
        )
        self.assertEqual(msg, str(exc.exception))

    def test_resolve_ranges(self):
        a_contract = RegistryPackageContract(
            package='fishtown-analytics-test/a',
            version='0.1.2'
        )
        b_contract = RegistryPackageContract(
            package='fishtown-analytics-test/a',
            version='<0.1.4'
        )
        a = RegistryUnpinnedPackage.from_contract(a_contract)
        b = RegistryUnpinnedPackage.from_contract(b_contract)
        c = a.incorporate(b)

        self.assertEqual(c.package, 'fishtown-analytics-test/a')
        self.assertEqual(
            c.versions,
            [
                VersionSpecifier(
                    build=None,
                    major='0',
                    matcher='=',
                    minor='1',
                    patch='2',
                    prerelease=None,
                ),
                VersionSpecifier(
                    build=None,
                    major='0',
                    matcher='<',
                    minor='1',
                    patch='4',
                    prerelease=None,
                ),
            ]
        )

        c_pinned = c.resolved()
        self.assertEqual(c_pinned.package, 'fishtown-analytics-test/a')
        self.assertEqual(c_pinned.version, '0.1.2')
        self.assertEqual(c_pinned.source_type(), 'hub')


class MockRegistry:
    def __init__(self, packages):
        self.packages = packages

    def index_cached(self, registry_base_url=None):
        return sorted(self.packages)

    def get_available_versions(self, name):
        try:
            pkg = self.packages[name]
        except KeyError:
            return []
        return list(pkg)

    def package_version(self, name, version):
        try:
            return self.packages[name][version]
        except KeyError:
            return None


class TestPackageSpec(unittest.TestCase):
    def setUp(self):
        self.patcher = mock.patch('dbt.task.deps.registry')
        self.registry = self.patcher.start()
        self.mock_registry = MockRegistry(packages={
            'fishtown-analytics-test/a': {
                '0.1.2': {
                    'id': 'fishtown-analytics-test/a/0.1.2',
                    'name': 'a',
                    'version': '0.1.2',
                    'packages': [],
                    '_source': {
                        'blahblah': 'asdfas',
                    },
                    'downloads': {
                        'tarball': 'https://example.com/invalid-url!',
                        'extra': 'field',
                    },
                    'newfield': ['another', 'value'],
                },
                '0.1.3': {
                    'id': 'fishtown-analytics-test/a/0.1.3',
                    'name': 'a',
                    'version': '0.1.3',
                    'packages': [],
                    '_source': {
                        'blahblah': 'asdfas',
                    },
                    'downloads': {
                        'tarball': 'https://example.com/invalid-url!',
                        'extra': 'field',
                    },
                    'newfield': ['another', 'value'],
                }
            },
            'fishtown-analytics-test/b': {
                '0.2.1': {
                    'id': 'fishtown-analytics-test/b/0.2.1',
                    'name': 'b',
                    'version': '0.2.1',
                    'packages': [{'package': 'fishtown-analytics-test/a', 'version': '>=0.1.3'}],
                    '_source': {
                        'blahblah': 'asdfas',
                    },
                    'downloads': {
                        'tarball': 'https://example.com/invalid-url!',
                        'extra': 'field',
                    },
                    'newfield': ['another', 'value'],
                },
            }
        })

        self.registry.index_cached.side_effect = self.mock_registry.index_cached
        self.registry.get_available_versions.side_effect = self.mock_registry.get_available_versions
        self.registry.package_version.side_effect = self.mock_registry.package_version

    def tearDown(self):
        self.patcher.stop()

    def test_dependency_resolution(self):
        package_config = PackageConfig.from_dict({
            'packages': [
                {'package': 'fishtown-analytics-test/a', 'version': '>0.1.2'},
                {'package': 'fishtown-analytics-test/b', 'version': '0.2.1'},
            ],
        })
        resolved = resolve_packages(package_config.packages, None)
        self.assertEqual(len(resolved), 2)
        self.assertEqual(resolved[0].name, 'fishtown-analytics-test/a')
        self.assertEqual(resolved[0].version, '0.1.3')
        self.assertEqual(resolved[1].name, 'fishtown-analytics-test/b')
        self.assertEqual(resolved[1].version, '0.2.1')



import os
from unittest import TestCase, mock

import dbt.flags
from dbt.node_types import NodeType
from dbt.source_config import SourceConfig

from .utils import config_from_parts_or_dicts


class SourceConfigTest(TestCase):
    def setUp(self):
        dbt.flags.STRICT_MODE = True
        dbt.flags.WARN_ERROR = True

        self.maxDiff = None

        profile_data = {
            'target': 'test',
            'quoting': {},
            'outputs': {
                'test': {
                    'type': 'redshift',
                    'host': 'localhost',
                    'schema': 'analytics',
                    'user': 'test',
                    'pass': 'test',
                    'dbname': 'test',
                    'port': 1,
                }
            }
        }

        root_project = {
            'name': 'root',
            'version': '0.1',
            'profile': 'test',
            'project-root': os.path.abspath('.'),
        }

        self.root_project_config = config_from_parts_or_dicts(
            project=root_project,
            profile=profile_data,
            cli_vars='{"test_schema_name": "foo"}'
        )

        snowplow_project = {
            'name': 'snowplow',
            'version': '0.1',
            'profile': 'test',
            'project-root': os.path.abspath('./dbt_modules/snowplow'),
        }

        self.snowplow_project_config = config_from_parts_or_dicts(
            project=snowplow_project, profile=profile_data
        )

        self.all_projects = {
            'root': self.root_project_config,
            'snowplow': self.snowplow_project_config
        }
        self.patcher = mock.patch('dbt.context.parser.get_adapter')
        self.factory = self.patcher.start()

    def tearDown(self):
        self.patcher.stop()

    def test__source_config_single_call(self):
        cfg = SourceConfig(self.root_project_config, self.root_project_config,
                           ['root', 'x'], NodeType.Model)
        cfg.update_in_model_config({
            'materialized': 'something',
            'sort': 'my sort key',
            'pre-hook': 'my pre run hook',
            'vars': {'a': 1, 'b': 2},
        })
        expect = {
            'column_types': {},
            'enabled': True,
            'materialized': 'something',
            'post-hook': [],
            'pre-hook': ['my pre run hook'],
            'persist_docs': {},
            'quoting': {},
            'sort': 'my sort key',
            'tags': [],
            'vars': {'a': 1, 'b': 2},
        }
        self.assertEqual(cfg.config, expect)

    def test__source_config_multiple_calls(self):
        cfg = SourceConfig(self.root_project_config, self.root_project_config,
                           ['root', 'x'], NodeType.Model)
        cfg.update_in_model_config({
            'materialized': 'something',
            'sort': 'my sort key',
            'pre-hook': 'my pre run hook',
            'vars': {'a': 1, 'b': 2},
        })
        cfg.update_in_model_config({
            'materialized': 'something else',
            'pre-hook': ['my other pre run hook', 'another pre run hook'],
            'vars': {'a': 4, 'c': 3},
        })
        expect = {
            'column_types': {},
            'enabled': True,
            'materialized': 'something else',
            'persist_docs': {},
            'post-hook': [],
            'pre-hook': [
                'my pre run hook',
                'my other pre run hook',
                'another pre run hook',
            ],
            'quoting': {},
            'sort': 'my sort key',
            'tags': [],
            'vars': {'a': 4, 'b': 2, 'c': 3},
        }
        self.assertEqual(cfg.config, expect)

    def test_source_config_all_keys_accounted_for(self):
        used_keys = frozenset(SourceConfig.AppendListFields) | \
                    frozenset(SourceConfig.ExtendDictFields) | \
                    frozenset(SourceConfig.ClobberFields)

        self.assertEqual(used_keys, frozenset(SourceConfig.ConfigKeys))

    def test__source_config_wrong_type(self):
        # ExtendDict fields should handle non-dict inputs gracefully
        self.root_project_config.models = {'persist_docs': False}
        cfg = SourceConfig(self.root_project_config, self.root_project_config,
                           ['root', 'x'], NodeType.Model)

        with self.assertRaises(dbt.exceptions.CompilationException) as exc:
            cfg.get_project_config(self.root_project_config)

        self.assertIn('must be a dict', str(exc.exception))



import unittest
from unittest import mock

import copy
from datetime import datetime

import dbt.flags
from dbt import tracking
from dbt.contracts.graph.manifest import Manifest, ManifestMetadata
from dbt.contracts.graph.parsed import (
    ParsedModelNode, DependsOn, NodeConfig, ParsedSeedNode
)
from dbt.contracts.graph.compiled import CompiledModelNode
from dbt.node_types import NodeType
import freezegun


REQUIRED_PARSED_NODE_KEYS = frozenset({
    'alias', 'tags', 'config', 'unique_id', 'refs', 'sources',
    'depends_on', 'database', 'schema', 'name', 'resource_type',
    'package_name', 'root_path', 'path', 'original_file_path', 'raw_sql',
    'docrefs', 'description', 'columns', 'fqn', 'build_path', 'patch_path',
})

REQUIRED_COMPILED_NODE_KEYS = frozenset(REQUIRED_PARSED_NODE_KEYS | {
    'compiled', 'extra_ctes_injected', 'extra_ctes', 'compiled_sql',
    'injected_sql', 'wrapped_sql'
})


class ManifestTest(unittest.TestCase):
    def setUp(self):
        dbt.flags.STRICT_MODE = True

        self.maxDiff = None

        self.model_config = NodeConfig.from_dict({
            'enabled': True,
            'materialized': 'view',
            'persist_docs': {},
            'post-hook': [],
            'pre-hook': [],
            'vars': {},
            'quoting': {},
            'column_types': {},
            'tags': [],
        })

        self.nested_nodes = {
            'model.snowplow.events': ParsedModelNode(
                name='events',
                database='dbt',
                schema='analytics',
                alias='events',
                resource_type=NodeType.Model,
                unique_id='model.snowplow.events',
                fqn=['snowplow', 'events'],
                package_name='snowplow',
                refs=[],
                sources=[],
                depends_on=DependsOn(),
                config=self.model_config,
                tags=[],
                path='events.sql',
                original_file_path='events.sql',
                root_path='',
                raw_sql='does not matter'
            ),
            'model.root.events': ParsedModelNode(
                name='events',
                database='dbt',
                schema='analytics',
                alias='events',
                resource_type=NodeType.Model,
                unique_id='model.root.events',
                fqn=['root', 'events'],
                package_name='root',
                refs=[],
                sources=[],
                depends_on=DependsOn(),
                config=self.model_config,
                tags=[],
                path='events.sql',
                original_file_path='events.sql',
                root_path='',
                raw_sql='does not matter'
            ),
            'model.root.dep': ParsedModelNode(
                name='dep',
                database='dbt',
                schema='analytics',
                alias='dep',
                resource_type=NodeType.Model,
                unique_id='model.root.dep',
                fqn=['root', 'dep'],
                package_name='root',
                refs=[['events']],
                sources=[],
                depends_on=DependsOn(nodes=['model.root.events']),
                config=self.model_config,
                tags=[],
                path='multi.sql',
                original_file_path='multi.sql',
                root_path='',
                raw_sql='does not matter'
            ),
            'model.root.nested': ParsedModelNode(
                name='nested',
                database='dbt',
                schema='analytics',
                alias='nested',
                resource_type=NodeType.Model,
                unique_id='model.root.nested',
                fqn=['root', 'nested'],
                package_name='root',
                refs=[['events']],
                sources=[],
                depends_on=DependsOn(nodes=['model.root.dep']),
                config=self.model_config,
                tags=[],
                path='multi.sql',
                original_file_path='multi.sql',
                root_path='',
                raw_sql='does not matter'
            ),
            'model.root.sibling': ParsedModelNode(
                name='sibling',
                database='dbt',
                schema='analytics',
                alias='sibling',
                resource_type=NodeType.Model,
                unique_id='model.root.sibling',
                fqn=['root', 'sibling'],
                package_name='root',
                refs=[['events']],
                sources=[],
                depends_on=DependsOn(nodes=['model.root.events']),
                config=self.model_config,
                tags=[],
                path='multi.sql',
                original_file_path='multi.sql',
                root_path='',
                raw_sql='does not matter'
            ),
            'model.root.multi': ParsedModelNode(
                name='multi',
                database='dbt',
                schema='analytics',
                alias='multi',
                resource_type=NodeType.Model,
                unique_id='model.root.multi',
                fqn=['root', 'multi'],
                package_name='root',
                refs=[['events']],
                sources=[],
                depends_on=DependsOn(nodes=['model.root.nested', 'model.root.sibling']),
                config=self.model_config,
                tags=[],
                path='multi.sql',
                original_file_path='multi.sql',
                root_path='',
                raw_sql='does not matter'
            ),
        }
        for node in self.nested_nodes.values():
            node.validate(node.to_dict())

    @freezegun.freeze_time('2018-02-14T09:15:13Z')
    def test__no_nodes(self):
        manifest = Manifest(nodes={}, macros={}, docs={},
                            generated_at=datetime.utcnow(), disabled=[],
                            files={})
        self.assertEqual(
            manifest.writable_manifest().to_dict(),
            {
                'nodes': {},
                'macros': {},
                'parent_map': {},
                'child_map': {},
                'generated_at': '2018-02-14T09:15:13Z',
                'docs': {},
                'metadata': {},
                'disabled': [],
                'files': {},
            }
        )

    @freezegun.freeze_time('2018-02-14T09:15:13Z')
    def test__nested_nodes(self):
        nodes = copy.copy(self.nested_nodes)
        manifest = Manifest(nodes=nodes, macros={}, docs={},
                            generated_at=datetime.utcnow(), disabled=[],
                            files={})
        serialized = manifest.writable_manifest().to_dict()
        self.assertEqual(serialized['generated_at'], '2018-02-14T09:15:13Z')
        self.assertEqual(serialized['docs'], {})
        self.assertEqual(serialized['disabled'], [])
        self.assertEqual(serialized['files'], {})
        parent_map = serialized['parent_map']
        child_map = serialized['child_map']
        # make sure there aren't any extra/missing keys.
        self.assertEqual(set(parent_map), set(nodes))
        self.assertEqual(set(child_map), set(nodes))
        self.assertEqual(
            parent_map['model.root.sibling'],
            ['model.root.events']
        )
        self.assertEqual(
            parent_map['model.root.nested'],
            ['model.root.dep']
        )
        self.assertEqual(
            parent_map['model.root.dep'],
            ['model.root.events']
        )
        # order doesn't matter.
        self.assertEqual(
            set(parent_map['model.root.multi']),
            set(['model.root.nested', 'model.root.sibling'])
        )
        self.assertEqual(
            parent_map['model.root.events'],
            [],
        )
        self.assertEqual(
            parent_map['model.snowplow.events'],
            [],
        )

        self.assertEqual(
            child_map['model.root.sibling'],
            ['model.root.multi'],
        )
        self.assertEqual(
            child_map['model.root.nested'],
            ['model.root.multi'],
        )
        self.assertEqual(
            child_map['model.root.dep'],
            ['model.root.nested']
        )
        self.assertEqual(
            child_map['model.root.multi'],
            []
        )
        self.assertEqual(
            set(child_map['model.root.events']),
            set(['model.root.dep', 'model.root.sibling'])
        )
        self.assertEqual(
            child_map['model.snowplow.events'],
            []
        )

    def test__build_flat_graph(self):
        nodes = copy.copy(self.nested_nodes)
        manifest = Manifest(nodes=nodes, macros={}, docs={},
                            generated_at=datetime.utcnow(), disabled=[],
                            files={})
        manifest.build_flat_graph()
        flat_graph = manifest.flat_graph
        flat_nodes = flat_graph['nodes']
        self.assertEqual(set(flat_graph), set(['nodes']))
        self.assertEqual(set(flat_nodes), set(self.nested_nodes))
        for node in flat_nodes.values():
            self.assertEqual(frozenset(node), REQUIRED_PARSED_NODE_KEYS)

    @mock.patch.object(tracking, 'active_user')
    def test_metadata(self, mock_user):
        mock_user.id = 'cfc9500f-dc7f-4c83-9ea7-2c581c1b38cf'
        mock_user.do_not_track = True
        self.assertEqual(
            ManifestMetadata(project_id='098f6bcd4621d373cade4e832627b4f6'),
            ManifestMetadata(
                project_id='098f6bcd4621d373cade4e832627b4f6',
                user_id='cfc9500f-dc7f-4c83-9ea7-2c581c1b38cf',
                send_anonymous_usage_stats=False,
            )
        )

    @mock.patch.object(tracking, 'active_user')
    @freezegun.freeze_time('2018-02-14T09:15:13Z')
    def test_no_nodes_with_metadata(self, mock_user):
        mock_user.id = 'cfc9500f-dc7f-4c83-9ea7-2c581c1b38cf'
        mock_user.do_not_track = True
        metadata = ManifestMetadata(project_id='098f6bcd4621d373cade4e832627b4f6')
        manifest = Manifest(nodes={}, macros={}, docs={},
                            generated_at=datetime.utcnow(), disabled=[],
                            metadata=metadata, files={})

        self.assertEqual(
            manifest.writable_manifest().to_dict(),
            {
                'nodes': {},
                'macros': {},
                'parent_map': {},
                'child_map': {},
                'generated_at': '2018-02-14T09:15:13Z',
                'docs': {},
                'metadata': {
                    'project_id': '098f6bcd4621d373cade4e832627b4f6',
                    'user_id': 'cfc9500f-dc7f-4c83-9ea7-2c581c1b38cf',
                    'send_anonymous_usage_stats': False,
                },
                'disabled': [],
                'files': {},
            }
        )

    def test_get_resource_fqns_empty(self):
        manifest = Manifest(nodes={}, macros={}, docs={},
                            generated_at=datetime.utcnow(), disabled=[],
                            files={})
        self.assertEqual(manifest.get_resource_fqns(), {})

    def test_get_resource_fqns(self):
        nodes = copy.copy(self.nested_nodes)
        nodes['seed.root.seed'] = ParsedSeedNode(
            name='seed',
            database='dbt',
            schema='analytics',
            alias='seed',
            resource_type='seed',
            unique_id='seed.root.seed',
            fqn=['root', 'seed'],
            package_name='root',
            refs=[['events']],
            sources=[],
            depends_on=DependsOn(),
            config=self.model_config,
            tags=[],
            path='seed.csv',
            original_file_path='seed.csv',
            root_path='',
            raw_sql='-- csv --',
            seed_file_path='data/seed.csv'
        )
        manifest = Manifest(nodes=nodes, macros={}, docs={},
                            generated_at=datetime.utcnow(), disabled=[],
                            files={})
        expect = {
            'models': frozenset([
                ('snowplow', 'events'),
                ('root', 'events'),
                ('root', 'dep'),
                ('root', 'nested'),
                ('root', 'sibling'),
                ('root', 'multi'),
            ]),
            'seeds': frozenset([('root', 'seed')]),
        }
        resource_fqns = manifest.get_resource_fqns()
        self.assertEqual(resource_fqns, expect)


class MixedManifestTest(unittest.TestCase):
    def setUp(self):
        dbt.flags.STRICT_MODE = True

        self.maxDiff = None

        self.model_config = NodeConfig.from_dict({
            'enabled': True,
            'materialized': 'view',
            'persist_docs': {},
            'post-hook': [],
            'pre-hook': [],
            'vars': {},
            'quoting': {},
            'column_types': {},
            'tags': [],
        })

        self.nested_nodes = {
            'model.snowplow.events': CompiledModelNode(
                name='events',
                database='dbt',
                schema='analytics',
                alias='events',
                resource_type=NodeType.Model,
                unique_id='model.snowplow.events',
                fqn=['snowplow', 'events'],
                package_name='snowplow',
                refs=[],
                sources=[],
                depends_on=DependsOn(),
                config=self.model_config,
                tags=[],
                path='events.sql',
                original_file_path='events.sql',
                root_path='',
                raw_sql='does not matter',
                compiled=True,
                compiled_sql='also does not matter',
                extra_ctes_injected=True,
                injected_sql=None,
                extra_ctes=[]
            ),
            'model.root.events': CompiledModelNode(
                name='events',
                database='dbt',
                schema='analytics',
                alias='events',
                resource_type=NodeType.Model,
                unique_id='model.root.events',
                fqn=['root', 'events'],
                package_name='root',
                refs=[],
                sources=[],
                depends_on=DependsOn(),
                config=self.model_config,
                tags=[],
                path='events.sql',
                original_file_path='events.sql',
                root_path='',
                raw_sql='does not matter',
                compiled=True,
                compiled_sql='also does not matter',
                extra_ctes_injected=True,
                injected_sql='and this also does not matter',
                extra_ctes=[]
            ),
            'model.root.dep': ParsedModelNode(
                name='dep',
                database='dbt',
                schema='analytics',
                alias='dep',
                resource_type=NodeType.Model,
                unique_id='model.root.dep',
                fqn=['root', 'dep'],
                package_name='root',
                refs=[['events']],
                sources=[],
                depends_on=DependsOn(nodes=['model.root.events']),
                config=self.model_config,
                tags=[],
                path='multi.sql',
                original_file_path='multi.sql',
                root_path='',
                raw_sql='does not matter'
            ),
            'model.root.nested': ParsedModelNode(
                name='nested',
                database='dbt',
                schema='analytics',
                alias='nested',
                resource_type=NodeType.Model,
                unique_id='model.root.nested',
                fqn=['root', 'nested'],
                package_name='root',
                refs=[['events']],
                sources=[],
                depends_on=DependsOn(nodes=['model.root.dep']),
                config=self.model_config,
                tags=[],
                path='multi.sql',
                original_file_path='multi.sql',
                root_path='',
                raw_sql='does not matter'
            ),
            'model.root.sibling': ParsedModelNode(
                name='sibling',
                database='dbt',
                schema='analytics',
                alias='sibling',
                resource_type=NodeType.Model,
                unique_id='model.root.sibling',
                fqn=['root', 'sibling'],
                package_name='root',
                refs=[['events']],
                sources=[],
                depends_on=DependsOn(nodes=['model.root.events']),
                config=self.model_config,
                tags=[],
                path='multi.sql',
                original_file_path='multi.sql',
                root_path='',
                raw_sql='does not matter'
            ),
            'model.root.multi': ParsedModelNode(
                name='multi',
                database='dbt',
                schema='analytics',
                alias='multi',
                resource_type=NodeType.Model,
                unique_id='model.root.multi',
                fqn=['root', 'multi'],
                package_name='root',
                refs=[['events']],
                sources=[],
                depends_on=DependsOn(nodes=['model.root.nested', 'model.root.sibling']),
                config=self.model_config,
                tags=[],
                path='multi.sql',
                original_file_path='multi.sql',
                root_path='',
                raw_sql='does not matter'
            ),
        }

    @freezegun.freeze_time('2018-02-14T09:15:13Z')
    def test__no_nodes(self):
        manifest = Manifest(nodes={}, macros={}, docs={},
                            generated_at=datetime.utcnow(), disabled=[],
                            files={})
        self.assertEqual(
            manifest.writable_manifest().to_dict(),
            {
                'nodes': {},
                'macros': {},
                'parent_map': {},
                'child_map': {},
                'generated_at': '2018-02-14T09:15:13Z',
                'docs': {},
                'metadata': {},
                'disabled': [],
                'files': {},
            }
        )

    @freezegun.freeze_time('2018-02-14T09:15:13Z')
    def test__nested_nodes(self):
        nodes = copy.copy(self.nested_nodes)
        manifest = Manifest(nodes=nodes, macros={}, docs={},
                            generated_at=datetime.utcnow(), disabled=[],
                            files={})
        serialized = manifest.writable_manifest().to_dict()
        self.assertEqual(serialized['generated_at'], '2018-02-14T09:15:13Z')
        self.assertEqual(serialized['disabled'], [])
        parent_map = serialized['parent_map']
        child_map = serialized['child_map']
        # make sure there aren't any extra/missing keys.
        self.assertEqual(set(parent_map), set(nodes))
        self.assertEqual(set(child_map), set(nodes))
        self.assertEqual(
            parent_map['model.root.sibling'],
            ['model.root.events']
        )
        self.assertEqual(
            parent_map['model.root.nested'],
            ['model.root.dep']
        )
        self.assertEqual(
            parent_map['model.root.dep'],
            ['model.root.events']
        )
        # order doesn't matter.
        self.assertEqual(
            set(parent_map['model.root.multi']),
            set(['model.root.nested', 'model.root.sibling'])
        )
        self.assertEqual(
            parent_map['model.root.events'],
            [],
        )
        self.assertEqual(
            parent_map['model.snowplow.events'],
            [],
        )

        self.assertEqual(
            child_map['model.root.sibling'],
            ['model.root.multi'],
        )
        self.assertEqual(
            child_map['model.root.nested'],
            ['model.root.multi'],
        )
        self.assertEqual(
            child_map['model.root.dep'],
            ['model.root.nested']
        )
        self.assertEqual(
            child_map['model.root.multi'],
            []
        )
        self.assertEqual(
            set(child_map['model.root.events']),
            set(['model.root.dep', 'model.root.sibling'])
        )
        self.assertEqual(
            child_map['model.snowplow.events'],
            []
        )

    def test__build_flat_graph(self):
        nodes = copy.copy(self.nested_nodes)
        manifest = Manifest(nodes=nodes, macros={}, docs={},
                            generated_at=datetime.utcnow(), disabled=[],
                            files={})
        manifest.build_flat_graph()
        flat_graph = manifest.flat_graph
        flat_nodes = flat_graph['nodes']
        self.assertEqual(set(flat_graph), set(['nodes']))
        self.assertEqual(set(flat_nodes), set(self.nested_nodes))
        compiled_count = 0
        for node in flat_nodes.values():
            if node.get('compiled'):
                self.assertEqual(frozenset(node), REQUIRED_COMPILED_NODE_KEYS)
                compiled_count += 1
            else:
                self.assertEqual(frozenset(node), REQUIRED_PARSED_NODE_KEYS)
        self.assertEqual(compiled_count, 2)



import os
import unittest

from dbt.contracts.graph.manifest import SourceFile, FileHash, FilePath, Manifest
from dbt.contracts.graph.parsed import ParsedDocumentation
from dbt.node_types import NodeType
from dbt.parser import docs
from dbt.parser.results import ParseResult
from dbt.parser.search import FileBlock

from .utils import config_from_parts_or_dicts


SNOWPLOW_SESSIONS_DOCS = r'''
This table contains one record for every session recorded by Snowplow.
A session is itself comprised of pageviews that all occur within 30 minutes
of each other. If more than 30 minutes elapse between pageviews, then a
new session is created. Given the following pageviews:

| session_id | page_view_id | page_title |
| ---------- | ------------ | ---------- |
| abc        | 123          | Home       |
| abc        | 456          | About      |
| abc        | 789          | Home       |

The following sessions will be created:

| session_id | first_page_title | count_pageviews |
| ---------- | ---------------- | --------------- |
| abc        | 123              | 2               |
| abc        | 789              | 1               |
'''

SNOWPLOW_SESSIONS_SESSION_ID_DOCS = r'''
This column is the unique identifier for a Snowplow session. It is generated by
a cookie then expires after 30 minutes of inactivity.
'''

SNOWPLOW_SESSIONS_BLOCK = r'''
{{% docs snowplow_sessions %}}
{snowplow_sessions_docs}
{{% enddocs %}}
'''.format(
        snowplow_sessions_docs=SNOWPLOW_SESSIONS_DOCS
).strip()


SNOWPLOW_SESSIONS_SESSION_ID_BLOCK = r'''
{{% docs snowplow_sessions__session_id %}}
{snowplow_sessions_session_id_docs}
{{% enddocs %}}
'''.format(
    snowplow_sessions_session_id_docs=SNOWPLOW_SESSIONS_SESSION_ID_DOCS
).strip()


TEST_DOCUMENTATION_FILE = r'''
{sessions_block}

{session_id_block}
'''.format(
    sessions_block=SNOWPLOW_SESSIONS_BLOCK,
    session_id_block=SNOWPLOW_SESSIONS_SESSION_ID_BLOCK,
)


class DocumentationParserTest(unittest.TestCase):
    def setUp(self):
        if os.name == 'nt':
            self.root_path = 'C:\\test_root'
            self.subdir_path = 'C:\\test_root\\test_subdir'
            self.testfile_path = 'C:\\test_root\\test_subdir\\test_file.md'
        else:
            self.root_path = '/test_root'
            self.subdir_path = '/test_root/test_subdir'
            self.testfile_path = '/test_root/test_subdir/test_file.md'

        profile_data = {
            'outputs': {
                'test': {
                    'type': 'postgres',
                    'host': 'localhost',
                    'schema': 'analytics',
                    'user': 'test',
                    'pass': 'test',
                    'dbname': 'test',
                    'port': 1,
                }
            },
            'target': 'test',
        }
        root_project = {
            'name': 'root',
            'version': '0.1',
            'profile': 'test',
            'project-root': self.root_path,
        }

        subdir_project = {
            'name': 'some_package',
            'version': '0.1',
            'profile': 'test',
            'project-root': self.subdir_path,
            'quoting': {},
        }
        self.root_project_config = config_from_parts_or_dicts(
            project=root_project, profile=profile_data
        )
        self.subdir_project_config = config_from_parts_or_dicts(
            project=subdir_project, profile=profile_data
        )

    def _build_file(self, contents, relative_path) -> FileBlock:
        match = FilePath(
            relative_path=relative_path,
            project_root=self.root_path,
            searched_path=self.subdir_path,
        )
        source_file = SourceFile(path=match, checksum=FileHash.empty())
        source_file.contents = contents
        return FileBlock(file=source_file)

    def test_load_file(self):
        parser = docs.DocumentationParser(
            results=ParseResult.rpc(),
            root_project=self.root_project_config,
            project=self.subdir_project_config,
            macro_manifest=Manifest.from_macros())

        file_block = self._build_file(TEST_DOCUMENTATION_FILE, 'test_file.md')

        parser.parse_file(file_block)
        results = sorted(parser.results.docs.values(), key=lambda n: n.name)
        self.assertEqual(len(results), 2)
        for result in results:
            self.assertIsInstance(result, ParsedDocumentation)
            self.assertEqual(result.package_name, 'some_package')
            self.assertNotEqual(result.file_contents, TEST_DOCUMENTATION_FILE)
            self.assertEqual(result.original_file_path, self.testfile_path)
            self.assertEqual(result.root_path, self.subdir_path)
            self.assertEqual(result.resource_type, NodeType.Documentation)
            self.assertEqual(result.path, 'test_file.md')

        self.assertEqual(results[0].name, 'snowplow_sessions')
        self.assertEqual(results[0].file_contents, SNOWPLOW_SESSIONS_BLOCK)
        self.assertEqual(results[1].name, 'snowplow_sessions__session_id')
        self.assertEqual(results[1].file_contents, SNOWPLOW_SESSIONS_SESSION_ID_BLOCK)

    def test_load_file_extras(self):
        TEST_DOCUMENTATION_FILE + '{% model foo %}select 1 as id{% endmodel %}'

        parser = docs.DocumentationParser(
            results=ParseResult.rpc(),
            root_project=self.root_project_config,
            project=self.subdir_project_config,
            macro_manifest=Manifest.from_macros())

        file_block = self._build_file(TEST_DOCUMENTATION_FILE, 'test_file.md')

        parser.parse_file(file_block)
        results = sorted(parser.results.docs.values(), key=lambda n: n.name)
        self.assertEqual(len(results), 2)
        for result in results:
            self.assertIsInstance(result, ParsedDocumentation)
        self.assertEqual(results[0].name, 'snowplow_sessions')
        self.assertEqual(results[1].name, 'snowplow_sessions__session_id')



from datetime import datetime
from decimal import Decimal
from unittest import mock
import unittest

import dbt.flags
from dbt.task import generate


class GenerateTest(unittest.TestCase):
    def setUp(self):
        dbt.flags.STRICT_MODE = True
        self.maxDiff = None
        self.manifest = mock.MagicMock()
        self.patcher = mock.patch('dbt.task.generate.get_unique_id_mapping')
        self.mock_get_unique_id_mapping = self.patcher.start()

    def tearDown(self):
        self.patcher.stop()

    def map_uids(self, effects):
        results = {
            generate.CatalogKey(db, sch, tbl): [uid]
            for db, sch, tbl, uid in effects
        }
        self.mock_get_unique_id_mapping.return_value = results

    def generate_catalog_dict(self, columns):
        result = generate.CatalogResults(
            nodes=generate.Catalog(columns).make_unique_id_map(self.manifest),
            generated_at=datetime.utcnow(),
        )
        return result.to_dict(omit_none=False)['nodes']

    def test__unflatten_empty(self):
        columns = {}
        expected = {}
        self.map_uids([])

        result = self.generate_catalog_dict(columns)

        self.mock_get_unique_id_mapping.assert_called_once_with(self.manifest)
        self.assertEqual(result, expected)

    def test__unflatten_one_column(self):
        columns = [{
            'column_comment': None,
            'column_index': Decimal('1'),
            'column_name': 'id',
            'column_type': 'integer',
            'table_comment': None,
            'table_name': 'test_table',
            'table_schema': 'test_schema',
            'table_type': 'BASE TABLE',
            'table_database': 'test_database',
        }]
        expected = {
            'test.model.test_table': {
                'metadata': {
                    'owner': None,
                    'comment': None,
                    'name': 'test_table',
                    'type': 'BASE TABLE',
                    'schema': 'test_schema',
                    'database': 'test_database',
                },
                'columns': {
                    'id': {
                        'type': 'integer',
                        'comment': None,
                        'index': 1,
                        'name': 'id'
                    },
                },
                'stats': {
                    'has_stats': {
                        'id': 'has_stats',
                        'label': 'Has Stats?',
                        'value': False,
                        'description': 'Indicates whether there are statistics for this table',
                        'include': False,
                    },
                },
                'unique_id': 'test.model.test_table',
            },
        }
        self.map_uids([('test_database', 'test_schema', 'test_table', 'test.model.test_table')])

        result = self.generate_catalog_dict(columns)

        self.mock_get_unique_id_mapping.assert_called_once_with(self.manifest)
        self.assertEqual(result, expected)

    def test__unflatten_multiple_schemas_dbs(self):
        columns = [
            {
                'column_comment': None,
                'column_index': Decimal('1'),
                'column_name': 'id',
                'column_type': 'integer',
                'table_comment': None,
                'table_name': 'test_table',
                'table_schema': 'test_schema',
                'table_type': 'BASE TABLE',
                'table_database': 'test_database',
                'table_owner': None,
            },
            {
                'column_comment': None,
                'column_index': Decimal('2'),
                'column_name': 'name',
                'column_type': 'text',
                'table_comment': None,
                'table_name': 'test_table',
                'table_schema': 'test_schema',
                'table_type': 'BASE TABLE',
                'table_database': 'test_database',
                'table_owner': None,
            },
            {
                'column_comment': None,
                'column_index': Decimal('1'),
                'column_name': 'id',
                'column_type': 'integer',
                'table_comment': None,
                'table_name': 'other_test_table',
                'table_schema': 'test_schema',
                'table_type': 'BASE TABLE',
                'table_database': 'test_database',
                'table_owner': None,
            },
            {
                'column_comment': None,
                'column_index': Decimal('2'),
                'column_name': 'email',
                'column_type': 'character varying',
                'table_comment': None,
                'table_name': 'other_test_table',
                'table_schema': 'test_schema',
                'table_type': 'BASE TABLE',
                'table_database': 'test_database',
                'table_owner': None,
            },
            {
                'column_comment': None,
                'column_index': Decimal('1'),
                'column_name': 'id',
                'column_type': 'integer',
                'table_comment': None,
                'table_name': 'test_table',
                'table_schema': 'other_test_schema',
                'table_type': 'BASE TABLE',
                'table_database': 'test_database',
                'table_owner': None,
            },
            {
                'column_comment': None,
                'column_index': Decimal('2'),
                'column_name': 'name',
                'column_type': 'text',
                'table_comment': None,
                'table_name': 'test_table',
                'table_schema': 'other_test_schema',
                'table_type': 'BASE TABLE',
                'table_database': 'test_database',
                'table_owner': None,
            },
            {
                'column_comment': None,
                'column_index': Decimal('1'),
                'column_name': 'id',
                'column_type': 'integer',
                'table_comment': None,
                'table_name': 'test_table',
                'table_schema': 'test_schema',
                'table_type': 'BASE TABLE',
                'table_database': 'other_test_database',
                'table_owner': None,
            },
            {
                'column_comment': None,
                'column_index': Decimal('2'),
                'column_name': 'name',
                'column_type': 'text',
                'table_comment': None,
                'table_name': 'test_table',
                'table_schema': 'test_schema',
                'table_type': 'BASE TABLE',
                'table_database': 'other_test_database',
                'table_owner': None,
            },
        ]
        expected = {
            'test.model.test_table': {
                'metadata': {
                    'owner': None,
                    'comment': None,
                    'name': 'test_table',
                    'type': 'BASE TABLE',
                    'schema': 'test_schema',
                    'database': 'test_database',
                },
                'columns': {
                    'id': {
                        'type': 'integer',
                        'comment': None,
                        'index': 1,
                        'name': 'id'
                    },
                    'name': {
                        'type': 'text',
                        'comment': None,
                        'index': 2,
                        'name': 'name',
                    }
                },
                'stats': {
                    'has_stats': {
                        'id': 'has_stats',
                        'label': 'Has Stats?',
                        'value': False,
                        'description': 'Indicates whether there are statistics for this table',
                        'include': False,
                    },
                },
                'unique_id': 'test.model.test_table',
            },
            'test.model.other_test_table': {
                'metadata': {
                    'owner': None,
                    'comment': None,
                    'name': 'other_test_table',
                    'type': 'BASE TABLE',
                    'schema': 'test_schema',
                    'database': 'test_database',
                },
                'columns': {
                    'id': {
                        'type': 'integer',
                        'comment': None,
                        'index': 1,
                        'name': 'id'
                    },
                    'email': {
                        'type': 'character varying',
                        'comment': None,
                        'index': 2,
                        'name': 'email',
                    }
                },
                'stats': {
                    'has_stats': {
                        'id': 'has_stats',
                        'label': 'Has Stats?',
                        'value': False,
                        'description': 'Indicates whether there are statistics for this table',
                        'include': False,
                    },
                },
                'unique_id': 'test.model.other_test_table',
            },
            'test.model.test_table_otherschema': {
                'metadata': {
                    'owner': None,
                    'comment': None,
                    'name': 'test_table',
                    'type': 'BASE TABLE',
                    'schema': 'other_test_schema',
                    'database': 'test_database',
                },
                'columns': {
                    'id': {
                        'type': 'integer',
                        'comment': None,
                        'index': 1,
                        'name': 'id'
                    },
                    'name': {
                        'type': 'text',
                        'comment': None,
                        'index': 2,
                        'name': 'name',
                    }
                },
                'stats': {
                    'has_stats': {
                        'id': 'has_stats',
                        'label': 'Has Stats?',
                        'value': False,
                        'description': 'Indicates whether there are statistics for this table',
                        'include': False,
                    },
                },
                'unique_id': 'test.model.test_table_otherschema',
            },
            'test.model.test_table_otherdb': {
                'metadata': {
                    'owner': None,
                    'comment': None,
                    'name': 'test_table',
                    'type': 'BASE TABLE',
                    'schema': 'test_schema',
                    'database': 'other_test_database',
                },
                'columns': {
                    'id': {
                        'type': 'integer',
                        'comment': None,
                        'index': 1,
                        'name': 'id'
                    },
                    'name': {
                        'type': 'text',
                        'comment': None,
                        'index': 2,
                        'name': 'name',
                    }
                },
                'stats': {
                    'has_stats': {
                        'id': 'has_stats',
                        'label': 'Has Stats?',
                        'value': False,
                        'description': 'Indicates whether there are statistics for this table',
                        'include': False,
                    },
                },
                'unique_id': 'test.model.test_table_otherdb',
            }
        }
        self.map_uids([
            (
                'test_database', 'test_schema', 'test_table',
                'test.model.test_table'
            ),
            (
                'test_database', 'test_schema', 'other_test_table',
                'test.model.other_test_table'
            ),
            (
                'test_database', 'other_test_schema', 'test_table',
                'test.model.test_table_otherschema'
            ),
            (
                'other_test_database', 'test_schema', 'test_table',
                'test.model.test_table_otherdb'
            ),
        ])

        result = self.generate_catalog_dict(columns)

        self.mock_get_unique_id_mapping.assert_called_once_with(self.manifest)
        self.assertEqual(result, expected)



import unittest

import decimal

from dbt.adapters.base import Column

class TestStringType(unittest.TestCase):

    def test__character_type(self):
        col = Column(
            'fieldname',
            'character',
            char_size=10
        )

        self.assertEqual(col.data_type, 'character varying(10)')


class TestNumericType(unittest.TestCase):

    def test__numeric_type(self):
        col = Column(
            'fieldname',
            'numeric',
            numeric_precision=decimal.Decimal('12'),
            numeric_scale=decimal.Decimal('2'))

        self.assertEqual(col.data_type, 'numeric(12,2)')

    def test__numeric_type_with_no_precision(self):
        # PostgreSQL, at least, will allow empty numeric precision
        col = Column(
            'fieldname',
            'numeric',
            numeric_precision=None)

        self.assertEqual(col.data_type, 'numeric')



import unittest
from unittest import mock

from .utils import config_from_parts_or_dicts, normalize

from dbt.contracts.graph.manifest import FileHash, FilePath, SourceFile
from dbt.parser import ParseResult
from dbt.parser.search import FileBlock
from dbt.parser import manifest


class MatchingHash(FileHash):
    def __init__(self):
        return super().__init__('', '')

    def __eq__(self, other):
        return True


class MismatchedHash(FileHash):
    def __init__(self):
        return super().__init__('', '')

    def __eq__(self, other):
        return False


class TestLoader(unittest.TestCase):
    def setUp(self):
        profile_data = {
            'target': 'test',
            'quoting': {},
            'outputs': {
                'test': {
                    'type': 'redshift',
                    'host': 'localhost',
                    'schema': 'analytics',
                    'user': 'test',
                    'pass': 'test',
                    'dbname': 'test',
                    'port': 1,
                }
            }
        }

        root_project = {
            'name': 'root',
            'version': '0.1',
            'profile': 'test',
            'project-root': normalize('/usr/src/app'),
        }

        self.root_project_config = config_from_parts_or_dicts(
            project=root_project,
            profile=profile_data,
            cli_vars='{"test_schema_name": "foo"}'
        )
        self.parser = mock.MagicMock()
        self.patched_result_builder = mock.patch('dbt.parser.manifest.make_parse_result')
        self.mock_result_builder = self.patched_result_builder.start()
        self.patched_result_builder.return_value = self._new_results()
        self.loader = manifest.ManifestLoader(
            self.root_project_config,
            {'root': self.root_project_config}
        )

    def _new_results(self):
        return ParseResult(MatchingHash(), MatchingHash(), {})

    def _mismatched_file(self, searched, name):
        return self._new_file(searched, name, False)

    def _matching_file(self, searched, name):
        return self._new_file(searched, name, True)

    def _new_file(self, searched, name, match):
        if match:
            checksum = MatchingHash()
        else:
            checksum = MismatchedHash()
        path = FilePath(
            searched_path=normalize(searched),
            relative_path=normalize(name),
            project_root=normalize(self.root_project_config.project_root),
        )
        return SourceFile(path=path, checksum=checksum)

    def test_model_no_cache(self):
        source_file = self._matching_file('models', 'model_1.sql')
        self.parser.load_file.return_value = source_file

        old_results = None

        self.loader.parse_with_cache(source_file.path, self.parser, old_results)
        # there was nothing in the cache, so parse_file should get called
        # with a FileBlock that has the given source file in it
        self.parser.parse_file.assert_called_once_with(FileBlock(file=source_file))

    def test_model_cache_hit(self):
        source_file = self._matching_file('models', 'model_1.sql')
        self.parser.load_file.return_value = source_file

        source_file_dupe = self._matching_file('models', 'model_1.sql')
        source_file_dupe.nodes.append('model.root.model_1')

        old_results = self._new_results()
        old_results.files[source_file_dupe.path.search_key] = source_file_dupe
        old_results.nodes = {'model.root.model_1': mock.MagicMock()}

        self.loader.parse_with_cache(source_file.path, self.parser, old_results)
        # there was a cache hit, so parse_file should never have been called
        self.parser.parse_file.assert_not_called()

    def test_model_cache_mismatch_checksum(self):
        source_file = self._mismatched_file('models', 'model_1.sql')
        self.parser.load_file.return_value = source_file

        source_file_dupe = self._mismatched_file('models', 'model_1.sql')
        source_file_dupe.nodes.append('model.root.model_1')

        old_results = self._new_results()
        old_results.files[source_file_dupe.path.search_key] = source_file_dupe
        old_results.nodes = {'model.root.model_1': mock.MagicMock()}

        self.loader.parse_with_cache(source_file.path, self.parser, old_results)
        # there was a cache checksum mismatch, so parse_file should get called
        # with a FileBlock that has the given source file in it
        self.parser.parse_file.assert_called_once_with(FileBlock(file=source_file))

    def test_model_cache_missing_file(self):
        source_file = self._matching_file('models', 'model_1.sql')
        self.parser.load_file.return_value = source_file

        source_file_different = self._matching_file('models', 'model_2.sql')
        source_file_different.nodes.append('model.root.model_2')

        old_results = self._new_results()
        old_results.files[source_file_different.path.search_key] = source_file_different
        old_results.nodes = {'model.root.model_2': mock.MagicMock()}

        self.loader.parse_with_cache(source_file.path, self.parser, old_results)
        # the filename wasn't in the cache, so parse_file should get called
        # with a  FileBlock that has the given source file in it.
        self.parser.parse_file.assert_called_once_with(FileBlock(file=source_file))



import unittest
from unittest import mock

from dbt.contracts.graph.parsed import ParsedModelNode, NodeConfig, DependsOn
from dbt.context import parser, runtime
from dbt.node_types import NodeType
import dbt.exceptions
from .mock_adapter import adapter_factory


class TestVar(unittest.TestCase):
    def setUp(self):
        self.model = ParsedModelNode(
            alias='model_one',
            name='model_one',
            database='dbt',
            schema='analytics',
            resource_type=NodeType.Model,
            unique_id='model.root.model_one',
            fqn=['root', 'model_one'],
            package_name='root',
            original_file_path='model_one.sql',
            root_path='/usr/src/app',
            refs=[],
            sources=[],
            depends_on=DependsOn(),
            config=NodeConfig.from_dict({
                'enabled': True,
                'materialized': 'view',
                'persist_docs': {},
                'post-hook': [],
                'pre-hook': [],
                'vars': {},
                'quoting': {},
                'column_types': {},
                'tags': [],
            }),
            tags=[],
            path='model_one.sql',
            raw_sql='',
            description='',
            columns={}
        )
        self.context = mock.MagicMock()

    def test_var_default_something(self):
        var = runtime.Var(self.model, self.context, overrides={'foo': 'baz'})
        self.assertEqual(var('foo'), 'baz')
        self.assertEqual(var('foo', 'bar'), 'baz')

    def test_var_default_none(self):
        var = runtime.Var(self.model, self.context, overrides={'foo': None})
        self.assertEqual(var('foo'), None)
        self.assertEqual(var('foo', 'bar'), None)

    def test_var_not_defined(self):
        var = runtime.Var(self.model, self.context, overrides={})

        self.assertEqual(var('foo', 'bar'), 'bar')
        with self.assertRaises(dbt.exceptions.CompilationException):
            var('foo')

    def test_parser_var_default_something(self):
        var = parser.Var(self.model, self.context, overrides={'foo': 'baz'})
        self.assertEqual(var('foo'), 'baz')
        self.assertEqual(var('foo', 'bar'), 'baz')

    def test_parser_var_default_none(self):
        var = parser.Var(self.model, self.context, overrides={'foo': None})
        self.assertEqual(var('foo'), None)
        self.assertEqual(var('foo', 'bar'), None)

    def test_parser_var_not_defined(self):
        # at parse-time, we should not raise if we encounter a missing var
        # that way disabled models don't get parse errors
        var = parser.Var(self.model, self.context, overrides={})

        self.assertEqual(var('foo', 'bar'), 'bar')
        self.assertEqual(var('foo'), None)


class TestParseWrapper(unittest.TestCase):
    def setUp(self):
        self.mock_config = mock.MagicMock()
        adapter_class = adapter_factory()
        self.mock_adapter = adapter_class(self.mock_config)
        self.wrapper = parser.DatabaseWrapper(self.mock_adapter)
        self.responder = self.mock_adapter.responder

    def test_unwrapped_method(self):
        self.assertEqual(self.wrapper.quote('test_value'), '"test_value"')
        self.responder.quote.assert_called_once_with('test_value')

    def test_wrapped_method(self):
        found = self.wrapper.get_relation('database', 'schema', 'identifier')
        self.assertEqual(found, None)
        self.responder.get_relation.assert_not_called()


class TestRuntimeWrapper(unittest.TestCase):
    def setUp(self):
        self.mock_config = mock.MagicMock()
        self.mock_config.quoting = {'database': True, 'schema': True, 'identifier': True}
        adapter_class = adapter_factory()
        self.mock_adapter = adapter_class(self.mock_config)
        self.wrapper = runtime.DatabaseWrapper(self.mock_adapter)
        self.responder = self.mock_adapter.responder

    def test_unwrapped_method(self):
        # the 'quote' method isn't wrapped, we should get our expected inputs
        self.assertEqual(self.wrapper.quote('test_value'), '"test_value"')
        self.responder.quote.assert_called_once_with('test_value')

    def test_wrapped_method(self):
        rel = mock.MagicMock()
        rel.matches.return_value = True
        self.responder.list_relations_without_caching.return_value = [rel]

        found = self.wrapper.get_relation('database', 'schema', 'identifier')

        self.assertEqual(found, rel)
        # it gets called with an information schema relation as the first arg,
        # which is hard to mock.
        self.responder.list_relations_without_caching.assert_called_once_with(
            mock.ANY, 'schema'
        )



import unittest
from unittest import mock

import os
import yaml

import dbt.flags
import dbt.parser
from dbt.exceptions import CompilationException
from dbt.parser import (
    ModelParser, MacroParser, DataTestParser, SchemaParser, ParserUtils,
    ParseResult, SnapshotParser, AnalysisParser
)
from dbt.parser.search import FileBlock
from dbt.parser.schema_test_builders import YamlBlock

from dbt.node_types import NodeType
from dbt.contracts.graph.manifest import (
    Manifest, FilePath, SourceFile, FileHash
)
from dbt.contracts.graph.parsed import (
    ParsedModelNode, ParsedMacro, ParsedNodePatch, ParsedSourceDefinition,
    NodeConfig, DependsOn, ColumnInfo, ParsedTestNode, TestConfig,
    ParsedSnapshotNode, TimestampSnapshotConfig, SnapshotStrategy,
    ParsedAnalysisNode
)
from dbt.contracts.graph.unparsed import FreshnessThreshold, ExternalTable

from .utils import config_from_parts_or_dicts, normalize


def get_abs_os_path(unix_path):
    return normalize(os.path.abspath(unix_path))


class BaseParserTest(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        dbt.flags.STRICT_MODE = True
        dbt.flags.WARN_ERROR = True

        self.maxDiff = None

        profile_data = {
            'target': 'test',
            'quoting': {},
            'outputs': {
                'test': {
                    'type': 'redshift',
                    'host': 'localhost',
                    'schema': 'analytics',
                    'user': 'test',
                    'pass': 'test',
                    'dbname': 'test',
                    'port': 1,
                }
            }
        }

        root_project = {
            'name': 'root',
            'version': '0.1',
            'profile': 'test',
            'project-root': normalize('/usr/src/app'),
        }

        self.root_project_config = config_from_parts_or_dicts(
            project=root_project,
            profile=profile_data,
            cli_vars='{"test_schema_name": "foo"}'
        )

        snowplow_project = {
            'name': 'snowplow',
            'version': '0.1',
            'profile': 'test',
            'project-root': get_abs_os_path('./dbt_modules/snowplow'),
        }

        self.snowplow_project_config = config_from_parts_or_dicts(
            project=snowplow_project, profile=profile_data
        )

        self.all_projects = {
            'root': self.root_project_config,
            'snowplow': self.snowplow_project_config
        }
        self.patcher = mock.patch('dbt.context.parser.get_adapter')
        self.factory = self.patcher.start()
        self.patcher_cmn = mock.patch('dbt.context.common.get_adapter')
        self.factory_cmn = self.patcher_cmn.start()

        self.macro_manifest = Manifest.from_macros()

    def tearDown(self):
        self.patcher_cmn.stop()
        self.patcher.stop()

    def file_block_for(self, data: str, filename: str, searched: str):
        root_dir = get_abs_os_path('./dbt_modules/snowplow')
        filename = normalize(filename)
        path = FilePath(
            searched_path=searched,
            relative_path=filename,
            project_root=root_dir,
        )
        source_file = SourceFile(
            path=path,
            checksum=FileHash.from_contents(data),
        )
        source_file.contents = data
        return FileBlock(file=source_file)

    def assert_has_results_length(self, results, files=1, macros=0, nodes=0,
                                  sources=0, docs=0, patches=0, disabled=0):
        self.assertEqual(len(results.files), files)
        self.assertEqual(len(results.macros), macros)
        self.assertEqual(len(results.nodes), nodes)
        self.assertEqual(len(results.sources), sources)
        self.assertEqual(len(results.docs), docs)
        self.assertEqual(len(results.patches), patches)
        self.assertEqual(sum(len(v) for v in results.disabled.values()), disabled)


SINGLE_TABLE_SOURCE = '''
version: 2
sources:
    - name: my_source
      tables:
        - name: my_table
'''

SINGLE_TABLE_SOURCE_TESTS = '''
version: 2
sources:
    - name: my_source
      tables:
        - name: my_table
          description: A description of my table
          columns:
            - name: color
              tests:
                - not_null:
                    severity: WARN
                - accepted_values:
                    values: ['red', 'blue', 'green']
'''


SINGLE_TABLE_MODEL_TESTS = '''
version: 2
models:
    - name: my_model
      description: A description of my model
      columns:
        - name: color
          description: The color value
          tests:
            - not_null:
                severity: WARN
            - accepted_values:
                values: ['red', 'blue', 'green']
            - foreign_package.test_case:
                arg: 100
'''


class SchemaParserTest(BaseParserTest):
    def setUp(self):
        super().setUp()
        self.parser = SchemaParser(
            results=ParseResult.rpc(),
            project=self.snowplow_project_config,
            root_project=self.root_project_config,
            macro_manifest=self.macro_manifest,
        )

    def file_block_for(self, data, filename):
        return super().file_block_for(data, filename, 'models')

    def yaml_block_for(self, test_yml: str, filename: str):
        file_block = self.file_block_for(data=test_yml, filename=filename)
        return YamlBlock.from_file_block(
            src=file_block,
            data=yaml.safe_load(test_yml),
        )


class SchemaParserSourceTest(SchemaParserTest):
    def test__read_basic_source(self):
        block = self.yaml_block_for(SINGLE_TABLE_SOURCE, 'test_one.yml')
        self.assertEqual(len(list(self.parser.read_yaml_models(yaml=block))), 0)
        results = list(self.parser.read_yaml_sources(yaml=block))
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].source.name, 'my_source')
        self.assertEqual(results[0].table.name, 'my_table')
        self.assertEqual(results[0].table.description, '')
        self.assertEqual(len(results[0].tests), 0)
        self.assertEqual(len(results[0].columns), 0)

    def test__parse_basic_source(self):
        block = self.file_block_for(SINGLE_TABLE_SOURCE, 'test_one.yml')
        self.parser.parse_file(block)
        # self.parser.parse_yaml_sources(yaml_block=block)
        self.assert_has_results_length(self.parser.results, sources=1)
        src = list(self.parser.results.sources.values())[0]
        expected = ParsedSourceDefinition(
            package_name='snowplow',
            source_name='my_source',
            schema='my_source',
            name='my_table',
            loader='',
            freshness=FreshnessThreshold(),
            external=ExternalTable(),
            source_description='',
            identifier='my_table',
            fqn=['snowplow', 'my_source', 'my_table'],
            database='test',
            unique_id='source.snowplow.my_source.my_table',
            root_path=get_abs_os_path('./dbt_modules/snowplow'),
            path=normalize('models/test_one.yml'),
            original_file_path=normalize('models/test_one.yml'),
            resource_type=NodeType.Source,
        )
        self.assertEqual(src, expected)

    def test__read_basic_source_tests(self):
        block = self.yaml_block_for(SINGLE_TABLE_SOURCE_TESTS, 'test_one.yml')
        self.assertEqual(len(list(self.parser.read_yaml_models(yaml=block))), 0)
        results = list(self.parser.read_yaml_sources(yaml=block))
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].source.name, 'my_source')
        self.assertEqual(results[0].table.name, 'my_table')
        self.assertEqual(results[0].table.description, 'A description of my table')
        self.assertEqual(len(results[0].columns), 1)
        self.assertEqual(len(results[0].columns[0].tests), 2)
        self.assertEqual(len(results[0].tests), 0)

    def test__parse_basic_source_tests(self):
        block = self.file_block_for(SINGLE_TABLE_SOURCE_TESTS, 'test_one.yml')
        self.parser.parse_file(block)
        self.assertEqual(len(self.parser.results.nodes), 2)
        self.assertEqual(len(self.parser.results.sources), 1)
        self.assertEqual(len(self.parser.results.patches), 0)
        src = list(self.parser.results.sources.values())[0]
        self.assertEqual(src.source_name, 'my_source')
        self.assertEqual(src.schema, 'my_source')
        self.assertEqual(src.name, 'my_table')
        self.assertEqual(src.description, 'A description of my table')

        tests = sorted(self.parser.results.nodes.values(), key=lambda n: n.unique_id)

        self.assertEqual(tests[0].config.severity, 'ERROR')
        self.assertEqual(tests[0].tags, ['schema'])
        self.assertEqual(tests[0].sources, [['my_source', 'my_table']])
        self.assertEqual(tests[0].column_name, 'color')
        self.assertEqual(tests[0].fqn, ['snowplow', 'schema_test', tests[0].name])
        self.assertEqual(tests[1].config.severity, 'WARN')
        self.assertEqual(tests[1].tags, ['schema'])
        self.assertEqual(tests[1].sources, [['my_source', 'my_table']])
        self.assertEqual(tests[1].column_name, 'color')
        self.assertEqual(tests[1].fqn, ['snowplow', 'schema_test', tests[1].name])

        path = get_abs_os_path('./dbt_modules/snowplow/models/test_one.yml')
        self.assertIn(path, self.parser.results.files)
        self.assertEqual(sorted(self.parser.results.files[path].nodes),
                         [t.unique_id for t in tests])
        self.assertIn(path, self.parser.results.files)
        self.assertEqual(self.parser.results.files[path].sources,
                         ['source.snowplow.my_source.my_table'])


class SchemaParserModelsTest(SchemaParserTest):
    def test__read_basic_model_tests(self):
        block = self.yaml_block_for(SINGLE_TABLE_MODEL_TESTS, 'test_one.yml')
        self.assertEqual(len(list(self.parser.read_yaml_sources(yaml=block))), 0)
        results = list(self.parser.read_yaml_models(yaml=block))
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].name, 'my_model')
        self.assertEqual(len(results[0].columns), 1)
        self.assertEqual(len(results[0].columns[0].tests), 3)
        self.assertEqual(len(results[0].tests), 0)

    def test__parse_basic_model_tests(self):
        block = self.file_block_for(SINGLE_TABLE_MODEL_TESTS, 'test_one.yml')
        self.parser.parse_file(block)
        self.assert_has_results_length(self.parser.results, patches=1, nodes=3)

        patch = list(self.parser.results.patches.values())[0]
        self.assertEqual(len(patch.columns), 1)
        self.assertEqual(patch.name, 'my_model')
        self.assertEqual(patch.description, 'A description of my model')
        expected_patch = ParsedNodePatch(
            name='my_model',
            description='A description of my model',
            columns={'color': ColumnInfo(name='color', description='The color value')},
            docrefs=[],
            original_file_path=normalize('models/test_one.yml'),
        )
        self.assertEqual(patch, expected_patch)

        tests = sorted(self.parser.results.nodes.values(), key=lambda n: n.unique_id)
        self.assertEqual(tests[0].config.severity, 'ERROR')
        self.assertEqual(tests[0].tags, ['schema'])
        self.assertEqual(tests[0].refs, [['my_model']])
        self.assertEqual(tests[0].column_name, 'color')
        self.assertEqual(tests[0].package_name, 'snowplow')
        self.assertTrue(tests[0].name.startswith('accepted_values_'))
        self.assertEqual(tests[0].fqn, ['snowplow', 'schema_test', tests[0].name])
        self.assertEqual(tests[0].unique_id.split('.'), ['test', 'snowplow', tests[0].name])
        self.assertEqual(tests[0].test_metadata.name, 'accepted_values')
        self.assertIsNone(tests[0].test_metadata.namespace)
        self.assertEqual(
            tests[0].test_metadata.kwargs,
            {
                'column_name': 'color',
                'values': ['red', 'blue', 'green'],
            }
        )

        # foreign packages are a bit weird, they include the macro package
        # name in the test name
        self.assertEqual(tests[1].config.severity, 'ERROR')
        self.assertEqual(tests[1].tags, ['schema'])
        self.assertEqual(tests[1].refs, [['my_model']])
        self.assertEqual(tests[1].column_name, 'color')
        self.assertEqual(tests[1].column_name, 'color')
        self.assertEqual(tests[1].fqn, ['snowplow', 'schema_test', tests[1].name])
        self.assertTrue(tests[1].name.startswith('foreign_package_test_case_'))
        self.assertEqual(tests[1].package_name, 'snowplow')
        self.assertEqual(tests[1].unique_id.split('.'), ['test', 'snowplow', tests[1].name])
        self.assertEqual(tests[1].test_metadata.name, 'test_case')
        self.assertEqual(tests[1].test_metadata.namespace, 'foreign_package')
        self.assertEqual(
            tests[1].test_metadata.kwargs,
            {
                'column_name': 'color',
                'arg': 100,
            },
        )

        self.assertEqual(tests[2].config.severity, 'WARN')
        self.assertEqual(tests[2].tags, ['schema'])
        self.assertEqual(tests[2].refs, [['my_model']])
        self.assertEqual(tests[2].column_name, 'color')
        self.assertEqual(tests[2].package_name, 'snowplow')
        self.assertTrue(tests[2].name.startswith('not_null_'))
        self.assertEqual(tests[2].fqn, ['snowplow', 'schema_test', tests[2].name])
        self.assertEqual(tests[2].unique_id.split('.'), ['test', 'snowplow', tests[2].name])
        self.assertEqual(tests[2].test_metadata.name, 'not_null')
        self.assertIsNone(tests[2].test_metadata.namespace)
        self.assertEqual(
            tests[2].test_metadata.kwargs,
            {
                'column_name': 'color',
            },
        )

        path = get_abs_os_path('./dbt_modules/snowplow/models/test_one.yml')
        self.assertIn(path, self.parser.results.files)
        self.assertEqual(sorted(self.parser.results.files[path].nodes),
                         [t.unique_id for t in tests])
        self.assertIn(path, self.parser.results.files)
        self.assertEqual(self.parser.results.files[path].patches, ['my_model'])


class ModelParserTest(BaseParserTest):
    def setUp(self):
        super().setUp()
        self.parser = ModelParser(
            results=ParseResult.rpc(),
            project=self.snowplow_project_config,
            root_project=self.root_project_config,
            macro_manifest=self.macro_manifest,
        )

    def file_block_for(self, data, filename):
        return super().file_block_for(data, filename, 'models')

    def test_basic(self):
        raw_sql = '{{ config(materialized="table") }}select 1 as id'
        block = self.file_block_for(raw_sql, 'nested/model_1.sql')
        self.parser.parse_file(block)
        self.assert_has_results_length(self.parser.results, nodes=1)
        node = list(self.parser.results.nodes.values())[0]
        expected = ParsedModelNode(
            alias='model_1',
            name='model_1',
            database='test',
            schema='analytics',
            resource_type=NodeType.Model,
            unique_id='model.snowplow.model_1',
            fqn=['snowplow', 'nested', 'model_1'],
            package_name='snowplow',
            original_file_path=normalize('models/nested/model_1.sql'),
            root_path=get_abs_os_path('./dbt_modules/snowplow'),
            config=NodeConfig(materialized='table'),
            path=normalize('nested/model_1.sql'),
            raw_sql=raw_sql,
        )
        self.assertEqual(node, expected)
        path = get_abs_os_path('./dbt_modules/snowplow/models/nested/model_1.sql')
        self.assertIn(path, self.parser.results.files)
        self.assertEqual(self.parser.results.files[path].nodes, ['model.snowplow.model_1'])

    def test_parse_error(self):
        block = self.file_block_for('{{ SYNTAX ERROR }}', 'nested/model_1.sql')
        with self.assertRaises(CompilationException):
            self.parser.parse_file(block)
        self.assert_has_results_length(self.parser.results, files=0)


class SnapshotParserTest(BaseParserTest):
    def setUp(self):
        super().setUp()
        self.parser = SnapshotParser(
            results=ParseResult.rpc(),
            project=self.snowplow_project_config,
            root_project=self.root_project_config,
            macro_manifest=self.macro_manifest,
        )

    def file_block_for(self, data, filename):
        return super().file_block_for(data, filename, 'snapshots')

    def test_parse_error(self):
        block = self.file_block_for('{% snapshot foo %}select 1 as id{%snapshot bar %}{% endsnapshot %}', 'nested/snap_1.sql')
        with self.assertRaises(CompilationException):
            self.parser.parse_file(block)
        self.assert_has_results_length(self.parser.results, files=0)

    def test_single_block(self):
        raw_sql = '''{{
                config(unique_key="id", target_schema="analytics",
                       target_database="dbt", strategy="timestamp",
                       updated_at="last_update")
            }}
            select 1 as id, now() as last_update'''
        full_file = '''
        {{% snapshot foo %}}{}{{% endsnapshot %}}
        '''.format(raw_sql)
        block = self.file_block_for(full_file, 'nested/snap_1.sql')
        self.parser.parse_file(block)
        self.assert_has_results_length(self.parser.results, nodes=1)
        node = list(self.parser.results.nodes.values())[0]
        expected = ParsedSnapshotNode(
            alias='foo',
            name='foo',
            # the `database` entry is overrridden by the target_database config
            database='dbt',
            schema='analytics',
            resource_type=NodeType.Snapshot,
            unique_id='snapshot.snowplow.foo',
            fqn=['snowplow', 'nested', 'snap_1', 'foo'],
            package_name='snowplow',
            original_file_path=normalize('snapshots/nested/snap_1.sql'),
            root_path=get_abs_os_path('./dbt_modules/snowplow'),
            config=TimestampSnapshotConfig(
                strategy=SnapshotStrategy.Timestamp,
                updated_at='last_update',
                target_database='dbt',
                target_schema='analytics',
                unique_key='id',
                materialized='snapshot',
            ),
            path=normalize('nested/snap_1.sql'),
            raw_sql=raw_sql,
        )
        self.assertEqual(node, expected)
        path = get_abs_os_path('./dbt_modules/snowplow/snapshots/nested/snap_1.sql')
        self.assertIn(path, self.parser.results.files)
        self.assertEqual(self.parser.results.files[path].nodes, ['snapshot.snowplow.foo'])

    def test_multi_block(self):
        raw_1 = '''
            {{
                config(unique_key="id", target_schema="analytics",
                       target_database="dbt", strategy="timestamp",
                       updated_at="last_update")
            }}
            select 1 as id, now() as last_update
        '''
        raw_2 = '''
            {{
                config(unique_key="id", target_schema="analytics",
                       target_database="dbt", strategy="timestamp",
                       updated_at="last_update")
            }}
            select 2 as id, now() as last_update
        '''
        full_file = '''
        {{% snapshot foo %}}{}{{% endsnapshot %}}
        {{% snapshot bar %}}{}{{% endsnapshot %}}
        '''.format(raw_1, raw_2)
        block = self.file_block_for(full_file, 'nested/snap_1.sql')
        self.parser.parse_file(block)
        self.assert_has_results_length(self.parser.results, nodes=2)
        nodes = sorted(self.parser.results.nodes.values(), key=lambda n: n.name)
        expect_foo = ParsedSnapshotNode(
            alias='foo',
            name='foo',
            database='dbt',
            schema='analytics',
            resource_type=NodeType.Snapshot,
            unique_id='snapshot.snowplow.foo',
            fqn=['snowplow', 'nested', 'snap_1', 'foo'],
            package_name='snowplow',
            original_file_path=normalize('snapshots/nested/snap_1.sql'),
            root_path=get_abs_os_path('./dbt_modules/snowplow'),
            config=TimestampSnapshotConfig(
                strategy=SnapshotStrategy.Timestamp,
                updated_at='last_update',
                target_database='dbt',
                target_schema='analytics',
                unique_key='id',
                materialized='snapshot',
            ),
            path=normalize('nested/snap_1.sql'),
            raw_sql=raw_1,
        )
        expect_bar = ParsedSnapshotNode(
            alias='bar',
            name='bar',
            database='dbt',
            schema='analytics',
            resource_type=NodeType.Snapshot,
            unique_id='snapshot.snowplow.bar',
            fqn=['snowplow', 'nested', 'snap_1', 'bar'],
            package_name='snowplow',
            original_file_path=normalize('snapshots/nested/snap_1.sql'),
            root_path=get_abs_os_path('./dbt_modules/snowplow'),
            config=TimestampSnapshotConfig(
                strategy=SnapshotStrategy.Timestamp,
                updated_at='last_update',
                target_database='dbt',
                target_schema='analytics',
                unique_key='id',
                materialized='snapshot',
            ),
            path=normalize('nested/snap_1.sql'),
            raw_sql=raw_2,
        )
        self.assertEqual(nodes[0], expect_bar)
        self.assertEqual(nodes[1], expect_foo)
        path = get_abs_os_path('./dbt_modules/snowplow/snapshots/nested/snap_1.sql')
        self.assertIn(path, self.parser.results.files)
        self.assertEqual(sorted(self.parser.results.files[path].nodes),
                         ['snapshot.snowplow.bar', 'snapshot.snowplow.foo'])


class MacroParserTest(BaseParserTest):
    def setUp(self):
        super().setUp()
        self.parser = MacroParser(
            results=ParseResult.rpc(),
            project=self.snowplow_project_config,
        )

    def file_block_for(self, data, filename):
        return super().file_block_for(data, filename, 'macros')

    def test_single_block(self):
        raw_sql = '{% macro foo(a, b) %}a ~ b{% endmacro %}'
        block = self.file_block_for(raw_sql, 'macro.sql')
        self.parser.parse_file(block)
        self.assert_has_results_length(self.parser.results, macros=1)
        macro = list(self.parser.results.macros.values())[0]
        expected = ParsedMacro(
            name='foo',
            resource_type=NodeType.Macro,
            unique_id='macro.snowplow.foo',
            package_name='snowplow',
            original_file_path=normalize('macros/macro.sql'),
            root_path=get_abs_os_path('./dbt_modules/snowplow'),
            path=normalize('macros/macro.sql'),
            raw_sql=raw_sql
        )
        self.assertEqual(macro, expected)
        path = get_abs_os_path('./dbt_modules/snowplow/macros/macro.sql')
        self.assertIn(path, self.parser.results.files)
        self.assertEqual(self.parser.results.files[path].macros, ['macro.snowplow.foo'])


class DataTestParserTest(BaseParserTest):
    def setUp(self):
        super().setUp()
        self.parser = DataTestParser(
            results=ParseResult.rpc(),
            project=self.snowplow_project_config,
            root_project=self.root_project_config,
            macro_manifest=self.macro_manifest,
        )

    def file_block_for(self, data, filename):
        return super().file_block_for(data, filename, 'tests')

    def test_basic(self):
        raw_sql = 'select * from {{ ref("blah") }} limit 0'
        block = self.file_block_for(raw_sql, 'test_1.sql')
        self.parser.parse_file(block)
        self.assert_has_results_length(self.parser.results, nodes=1)
        node = list(self.parser.results.nodes.values())[0]
        expected = ParsedTestNode(
            alias='test_1',
            name='test_1',
            database='test',
            schema='analytics',
            resource_type=NodeType.Test,
            unique_id='test.snowplow.test_1',
            fqn=['snowplow', 'data_test', 'test_1'],
            package_name='snowplow',
            original_file_path=normalize('tests/test_1.sql'),
            root_path=get_abs_os_path('./dbt_modules/snowplow'),
            refs=[['blah']],
            config=TestConfig(severity='ERROR'),
            tags=['data'],
            path=normalize('data_test/test_1.sql'),
            raw_sql=raw_sql,
        )
        self.assertEqual(node, expected)
        path = get_abs_os_path('./dbt_modules/snowplow/tests/test_1.sql')
        self.assertIn(path, self.parser.results.files)
        self.assertEqual(self.parser.results.files[path].nodes, ['test.snowplow.test_1'])


class AnalysisParserTest(BaseParserTest):
    def setUp(self):
        super().setUp()
        self.parser = AnalysisParser(
            results=ParseResult.rpc(),
            project=self.snowplow_project_config,
            root_project=self.root_project_config,
            macro_manifest=self.macro_manifest,
        )

    def file_block_for(self, data, filename):
        return super().file_block_for(data, filename, 'analyses')

    def test_basic(self):
        raw_sql = 'select 1 as id'
        block = self.file_block_for(raw_sql, 'nested/analysis_1.sql')
        self.parser.parse_file(block)
        self.assert_has_results_length(self.parser.results, nodes=1)
        node = list(self.parser.results.nodes.values())[0]
        expected = ParsedAnalysisNode(
            alias='analysis_1',
            name='analysis_1',
            database='test',
            schema='analytics',
            resource_type=NodeType.Analysis,
            unique_id='analysis.snowplow.analysis_1',
            fqn=['snowplow', 'analysis', 'nested', 'analysis_1'],
            package_name='snowplow',
            original_file_path=normalize('analyses/nested/analysis_1.sql'),
            root_path=get_abs_os_path('./dbt_modules/snowplow'),
            depends_on=DependsOn(),
            config=NodeConfig(),
            path=normalize('analysis/nested/analysis_1.sql'),
            raw_sql=raw_sql,
        )
        self.assertEqual(node, expected)
        path = get_abs_os_path('./dbt_modules/snowplow/analyses/nested/analysis_1.sql')
        self.assertIn(path, self.parser.results.files)
        self.assertEqual(self.parser.results.files[path].nodes, ['analysis.snowplow.analysis_1'])


class ParserUtilsTest(unittest.TestCase):
    def setUp(self):
        x_depends_on = mock.MagicMock()
        y_depends_on = mock.MagicMock()
        x_uid = 'model.project.x'
        y_uid = 'model.otherproject.y'
        src_uid = 'source.thirdproject.src.tbl'
        remote_docref = mock.MagicMock(documentation_package='otherproject', documentation_name='my_doc', column_name=None)
        docref = mock.MagicMock(documentation_package='', documentation_name='my_doc', column_name=None)
        self.x_node = mock.MagicMock(
            refs=[], sources=[['src', 'tbl']], docrefs=[remote_docref], unique_id=x_uid,
            resource_type=NodeType.Model, depends_on=x_depends_on,
            description='other_project: {{ doc("otherproject", "my_doc") }}',
        )
        self.y_node = mock.MagicMock(
            refs=[['x']], sources=[], docrefs=[docref], unique_id=y_uid,
            resource_type=NodeType.Model, depends_on=y_depends_on,
            description='{{ doc("my_doc") }}',
        )
        self.src_node = mock.MagicMock(
            resource_type=NodeType.Source, unique_id=src_uid,
        )
        nodes = {
            x_uid: self.x_node,
            y_uid: self.y_node,
            src_uid: self.src_node,
        }
        docs = {
            'otherproject.my_doc': mock.MagicMock(block_contents='some docs')
        }
        self.manifest = Manifest(
            nodes=nodes, macros={}, docs=docs, disabled=[], files={}, generated_at=mock.MagicMock()
        )

    def test_resolve_docs(self):
        # no error. TODO: real test
        result = ParserUtils.process_docs(self.manifest, 'project')
        self.assertIs(result, self.manifest)
        self.assertEqual(self.x_node.description, 'other_project: some docs')
        self.assertEqual(self.y_node.description, 'some docs')

    def test_resolve_sources(self):
        result = ParserUtils.process_sources(self.manifest, 'project')
        self.assertIs(result, self.manifest)
        self.x_node.depends_on.nodes.append.assert_called_once_with('source.thirdproject.src.tbl')

    def test_resolve_refs(self):
        result = ParserUtils.process_refs(self.manifest, 'project')
        self.assertIs(result, self.manifest)
        self.y_node.depends_on.nodes.append.assert_called_once_with('model.project.x')



import unittest
from unittest.mock import patch, MagicMock

import dbt.main
import dbt.version
import sys


class VersionTest(unittest.TestCase):

    @patch("dbt.version.__version__", "0.10.0")
    @patch('dbt.version.requests.get')
    def test_versions_equal(self, mock_get):
        mock_get.return_value.json.return_value = {'info': {'version': '0.10.0'}}

        latest_version = dbt.version.get_latest_version()
        installed_version = dbt.version.get_installed_version()
        version_information = dbt.version.get_version_information()

        expected_version_information = "installed version: 0.10.0\n" \
            "   latest version: 0.10.0\n\n" \
            "Up to date!"

        self.assertEqual(latest_version, installed_version)
        self.assertEqual(latest_version, installed_version)
        self.assertMultiLineEqual(version_information,
                                  expected_version_information)

    @patch("dbt.version.__version__", "0.10.2-a1")
    @patch('dbt.version.requests.get')
    def test_installed_version_greater(self, mock_get):
        mock_get.return_value.json.return_value = {'info': {'version': '0.10.1'}}

        latest_version = dbt.version.get_latest_version()
        installed_version = dbt.version.get_installed_version()
        version_information = dbt.version.get_version_information()

        expected_version_information = "installed version: 0.10.2-a1\n" \
            "   latest version: 0.10.1\n\n" \
            "Your version of dbt is ahead of the latest release!"

        assert installed_version > latest_version
        self.assertMultiLineEqual(version_information,
                                  expected_version_information)

    @patch("dbt.version.__version__", "0.9.5")
    @patch('dbt.version.requests.get')
    def test_installed_version_lower(self, mock_get):
        mock_get.return_value.json.return_value = {'info': {'version': '0.10.0'}}

        latest_version = dbt.version.get_latest_version()
        installed_version = dbt.version.get_installed_version()
        version_information = dbt.version.get_version_information()

        expected_version_information = "installed version: 0.9.5\n" \
            "   latest version: 0.10.0\n\n" \
            "Your version of dbt is out of date! " \
            "You can find instructions for upgrading here:\n" \
            "https://docs.getdbt.com/docs/installation"

        assert installed_version < latest_version
        self.assertMultiLineEqual(version_information,
                                  expected_version_information)

    # suppress having version info printed to the screen during tests.
    @patch('sys.stderr')
    @patch('dbt.version.requests.get')
    def test_dbt_version_flag(self, mock_get, stderr):
        mock_get.return_value.json.return_value = {'info': {'version': '0.10.1'}}

        with self.assertRaises(SystemExit) as exc:
            dbt.main.handle_and_check(['--version'])
        self.assertEqual(exc.exception.code, 0)



import unittest
from unittest import mock

import string
import dbt.exceptions
import dbt.graph.selector as graph_selector

import networkx as nx


class BaseGraphSelectionTest(unittest.TestCase):
    def create_graph(self):
        raise NotImplementedError

    def add_tags(self, nodes):
        pass

    def setUp(self):
        self.package_graph = graph_selector.Graph(self.create_graph())
        nodes = {
            node: mock.MagicMock(fqn=node.split('.')[1:], tags=[])
            for node in self.package_graph
        }
        self.add_tags(nodes)
        self.manifest = mock.MagicMock(nodes=nodes)
        self.selector = graph_selector.NodeSelector(self.package_graph, self.manifest)


class GraphSelectionTest(BaseGraphSelectionTest):
    def create_graph(self):
        integer_graph = nx.balanced_tree(2, 2, nx.DiGraph())

        package_mapping = {
            i: 'm.' + ('X' if i % 2 == 0 else 'Y') + '.' + letter
            for (i, letter) in enumerate(string.ascii_lowercase)
        }

        # Edges: [(X.a, Y.b), (X.a, X.c), (Y.b, Y.d), (Y.b, X.e), (X.c, Y.f), (X.c, X.g)]
        return nx.relabel_nodes(integer_graph, package_mapping)

    def add_tags(self, nodes):
        nodes['m.X.a'].tags = ['abc']
        nodes['m.Y.b'].tags = ['abc']
        nodes['m.X.c'].tags = ['abc']
        nodes['m.Y.d'].tags = []
        nodes['m.X.e'].tags = ['efg']
        nodes['m.Y.f'].tags = ['efg']
        nodes['m.X.g'].tags = ['efg']

    def run_specs_and_assert(self, graph, include, exclude, expected):
        selected = self.selector.select_nodes(
            graph,
            include,
            exclude
        )

        self.assertEqual(selected, expected)


    def test__single_node_selection_in_package(self):
        self.run_specs_and_assert(
            self.package_graph,
            ['X.a'],
            [],
            set(['m.X.a'])
        )

    def test__select_by_tag(self):
        self.run_specs_and_assert(
            self.package_graph,
            ['tag:abc'],
            [],
            set(['m.X.a', 'm.Y.b', 'm.X.c'])
        )

    def test__exclude_by_tag(self):
        self.run_specs_and_assert(
            self.package_graph,
            ['*'],
            ['tag:abc'],
            set(['m.Y.d', 'm.X.e', 'm.Y.f', 'm.X.g'])
        )

    def test__select_by_tag_and_model_name(self):
        self.run_specs_and_assert(
            self.package_graph,
            ['tag:abc', 'a'],
            [],
            set(['m.X.a', 'm.Y.b', 'm.X.c'])
        )

        self.run_specs_and_assert(
            self.package_graph,
            ['tag:abc', 'd'],
            [],
            set(['m.X.a', 'm.Y.b', 'm.X.c', 'm.Y.d'])
        )

    def test__multiple_node_selection_in_package(self):
        self.run_specs_and_assert(
            self.package_graph,
            ['X.a', 'b'],
            [],
            set(['m.X.a', 'm.Y.b'])
        )

    def test__select_children_except_in_package(self):
        self.run_specs_and_assert(
            self.package_graph,
            ['X.a+'],
            ['b'],
            set(['m.X.a','m.X.c', 'm.Y.d','m.X.e','m.Y.f','m.X.g']))

    def test__select_children_except_tag(self):
        self.run_specs_and_assert(
            self.package_graph,
            ['X.a+'],
            ['tag:efg'],
            set(['m.X.a','m.Y.b','m.X.c', 'm.Y.d']))

    def test__select_childrens_parents(self):
        self.run_specs_and_assert(
            self.package_graph,
            ['@X.c'],
            [],
            set(['m.X.a', 'm.X.c', 'm.Y.f', 'm.X.g'])
        )

    def parse_spec_and_assert(self, spec, parents, children, filter_type, filter_value, childrens_parents):
        parsed = graph_selector.SelectionCriteria(spec)
        self.assertEqual(parsed.select_parents, parents)
        self.assertEqual(parsed.select_children, children)
        self.assertEqual(parsed.selector_type, filter_type)
        self.assertEqual(parsed.selector_value, filter_value)
        self.assertEqual(parsed.select_childrens_parents, childrens_parents)

    def invalid_spec(self, spec):
        with self.assertRaises(dbt.exceptions.RuntimeException):
            graph_selector.SelectionCriteria(spec)

    def test__spec_parsing(self):
        self.parse_spec_and_assert('a', False, False, 'fqn', 'a', False)
        self.parse_spec_and_assert('+a', True, False, 'fqn', 'a', False)
        self.parse_spec_and_assert('a+', False, True, 'fqn', 'a', False)
        self.parse_spec_and_assert('+a+', True, True, 'fqn', 'a', False)
        self.parse_spec_and_assert('@a', False, False, 'fqn', 'a', True)
        self.invalid_spec('@a+')

        self.parse_spec_and_assert('a.b', False, False, 'fqn', 'a.b', False)
        self.parse_spec_and_assert('+a.b', True, False, 'fqn', 'a.b', False)
        self.parse_spec_and_assert('a.b+', False, True, 'fqn', 'a.b', False)
        self.parse_spec_and_assert('+a.b+', True, True, 'fqn', 'a.b', False)
        self.parse_spec_and_assert('@a.b', False, False, 'fqn', 'a.b', True)
        self.invalid_spec('@a.b+')

        self.parse_spec_and_assert('a.b.*', False, False, 'fqn', 'a.b.*', False)
        self.parse_spec_and_assert('+a.b.*', True, False, 'fqn', 'a.b.*', False)
        self.parse_spec_and_assert('a.b.*+', False, True, 'fqn', 'a.b.*', False)
        self.parse_spec_and_assert('+a.b.*+', True, True, 'fqn', 'a.b.*', False)
        self.parse_spec_and_assert('@a.b.*', False, False, 'fqn', 'a.b.*', True)
        self.invalid_spec('@a.b*+')

        self.parse_spec_and_assert('tag:a', False, False, 'tag', 'a', False)
        self.parse_spec_and_assert('+tag:a', True, False, 'tag', 'a', False)
        self.parse_spec_and_assert('tag:a+', False, True, 'tag', 'a', False)
        self.parse_spec_and_assert('+tag:a+', True, True, 'tag', 'a', False)
        self.parse_spec_and_assert('@tag:a', False, False, 'tag', 'a', True)
        self.invalid_spec('@tag:a+')

        self.parse_spec_and_assert('source:a', False, False, 'source', 'a', False)
        self.parse_spec_and_assert('source:a+', False, True, 'source', 'a', False)
        self.parse_spec_and_assert('@source:a', False, False, 'source', 'a', True)
        self.invalid_spec('@source:a+')

    def test__package_name_getter(self):
        found = graph_selector.get_package_names(self.package_graph)

        expected = set(['X', 'Y'])
        self.assertEqual(found, expected)

    def assert_is_selected_node(self, node, spec, should_work):
        self.assertEqual(
            graph_selector.is_selected_node(node, spec),
            should_work
        )

    def test__is_selected_node(self):
        self.assert_is_selected_node(('X', 'a'), ('a'), True)
        self.assert_is_selected_node(('X', 'a'), ('X', 'a'), True)
        self.assert_is_selected_node(('X', 'a'), ('*'), True)
        self.assert_is_selected_node(('X', 'a'), ('X', '*'), True)

        self.assert_is_selected_node(('X', 'a', 'b', 'c'), ('X', '*'), True)
        self.assert_is_selected_node(('X', 'a', 'b', 'c'), ('X', 'a', '*'), True)
        self.assert_is_selected_node(('X', 'a', 'b', 'c'), ('X', 'a', 'b', '*'), True)
        self.assert_is_selected_node(('X', 'a', 'b', 'c'), ('X', 'a', 'b', 'c'), True)
        self.assert_is_selected_node(('X', 'a', 'b', 'c'), ('X', 'a'), True)
        self.assert_is_selected_node(('X', 'a', 'b', 'c'), ('X', 'a', 'b'), True)

        self.assert_is_selected_node(('X', 'a'), ('b'), False)
        self.assert_is_selected_node(('X', 'a'), ('X', 'b'), False)
        self.assert_is_selected_node(('X', 'a'), ('X', 'a', 'b'), False)
        self.assert_is_selected_node(('X', 'a'), ('Y', '*'), False)



"""Unit test utility functions.

Note that all imports should be inside the functions to avoid import/mocking
issues.
"""
import os
from unittest import mock
from unittest import TestCase

from hologram import ValidationError


def normalize(path):
    """On windows, neither is enough on its own:

    >>> normcase('C:\\documents/ALL CAPS/subdir\\..')
    'c:\\documents\\all caps\\subdir\\..'
    >>> normpath('C:\\documents/ALL CAPS/subdir\\..')
    'C:\\documents\\ALL CAPS'
    >>> normpath(normcase('C:\\documents/ALL CAPS/subdir\\..'))
    'c:\\documents\\all caps'
    """
    return os.path.normcase(os.path.normpath(path))


class Obj:
    which = 'blah'


def mock_connection(name):
    conn = mock.MagicMock()
    conn.name = name
    return conn


def config_from_parts_or_dicts(project, profile, packages=None, cli_vars='{}'):
    from dbt.config import Project, Profile, RuntimeConfig
    from dbt.utils import parse_cli_vars
    from copy import deepcopy
    if not isinstance(cli_vars, dict):
        cli_vars = parse_cli_vars(cli_vars)
    if not isinstance(project, Project):
        project = Project.from_project_config(deepcopy(project), packages)
    if not isinstance(profile, Profile):
        profile = Profile.from_raw_profile_info(deepcopy(profile),
                                                project.profile_name,
                                                cli_vars)
    args = Obj()
    args.vars = repr(cli_vars)
    args.profile_dir = '/dev/null'
    return RuntimeConfig.from_parts(
        project=project,
        profile=profile,
        args=args
    )


def inject_adapter(value):
    """Inject the given adapter into the adapter factory, so your hand-crafted
    artisanal adapter will be available from get_adapter() as if dbt loaded it.
    """
    from dbt.adapters.factory import FACTORY
    key = value.type()
    FACTORY.adapters[key] = value
    FACTORY.adapter_types[key] = type(value)


class ContractTestCase(TestCase):
    ContractType = None

    def setUp(self):
        self.maxDiff = None
        super().setUp()

    def assert_to_dict(self, obj, dct):
        self.assertEqual(obj.to_dict(), dct)

    def assert_from_dict(self, obj, dct, cls=None):
        if cls is None:
            cls = self.ContractType
        self.assertEqual(cls.from_dict(dct),  obj)

    def assert_symmetric(self, obj, dct, cls=None):
        self.assert_to_dict(obj, dct)
        self.assert_from_dict(obj, dct, cls)

    def assert_fails_validation(self, dct, cls=None):
        if cls is None:
            cls = self.ContractType

        with self.assertRaises(ValidationError):
            cls.from_dict(dct)
